from safeds.data.image.typing import ImageSize


class DatasetMissesFeaturesError(ValueError):
    """
    Raised when a dataset misses feature columns.

    Parameters
    ----------
    missing_feature_names:
        The names of the missing feature columns.
    """

    def __init__(self, missing_feature_names: list[str]):
        super().__init__(f"Dataset misses the feature columns '{missing_feature_names}'.")


class DatasetMissesDataError(ValueError):
    """Raised when a dataset contains no rows."""

    def __init__(self) -> None:
        super().__init__("Dataset contains no rows")


class LearningError(Exception):
    """
    Raised when an error occurred while training a model.

    Parameters
    ----------
    reason:
        The reason for the error.
    """

    def __init__(self, reason: str):
        super().__init__(f"Error occurred while learning: {reason}")


class ModelNotFittedError(RuntimeError):
    """Raised when a model is used before fitting it."""

    def __init__(self) -> None:
        super().__init__("The model has not been fitted yet.")


class InvalidModelStructureError(Exception):
    """Raised when the structure of the model is invalid."""

    def __init__(self, reason: str) -> None:
        super().__init__(f"The model structure is invalid: {reason}")


class PredictionError(Exception):
    """
    Raised when an error occurred while prediction a target vector using a model.

    Parameters
    ----------
    reason:
        The reason for the error.
    """

    def __init__(self, reason: str):
        super().__init__(f"Error occurred while predicting: {reason}")


class FeatureDataMismatchError(Exception):
    """Raised when the columns of the table passed to the predict or fit method do not match with the specified features of the neural network."""

    def __init__(self) -> None:
        super().__init__(
            "The features in the given table do not match with the specified feature columns names of the neural network.",
        )


class InputSizeError(Exception):
    """Raised when the amount of features being passed to a network does not match with its input size."""

    def __init__(self, data_size: int | ImageSize, input_layer_size: int | ImageSize) -> None:
        super().__init__(
            f"The data size being passed to the network({data_size}) does not match with its input size({input_layer_size}). Consider changing the data size of the model or reformatting the data.",
        )


class PlainTableError(TypeError):
    """Raised when a plain table is used instead of a TabularDataset."""

    def __init__(self) -> None:
        super().__init__(
            (
                "This method needs a tabular dataset. "
                "It knows which columns are features and which are the target to predict.\n"
                "Use `Table.to_tabular_dataset()` to create a tabular dataset."
            ),
        )



"""Converters between our data contains and tensors."""

from typing import TYPE_CHECKING

import apipkg

if TYPE_CHECKING:
    from ._input_converter import InputConversion
    from ._input_converter_image import InputConversionImage
    from ._input_converter_table import InputConversionTable
    from ._input_converter_time_series import InputConversionTimeSeries
    from ._output_converter import OutputConversion
    from ._output_converter_image import (
        OutputConversionImageToColumn,
        OutputConversionImageToImage,
        OutputConversionImageToTable,
    )
    from ._output_converter_table import OutputConversionTable
    from ._output_converter_time_series import OutputConversionTimeSeries

apipkg.initpkg(
    __name__,
    {
        "InputConversion": "._input_converter:InputConversion",
        "InputConversionImage": "._input_converter_image:InputConversionImage",
        "InputConversionTable": "._input_converter_table:InputConversionTable",
        "InputConversionTimeSeries": "._input_converter_time_series:InputConversionTimeSeries",
        "OutputConversion": "._output_converter:OutputConversion",
        "OutputConversionImageToColumn": "._output_converter_image:OutputConversionImageToColumn",
        "OutputConversionImageToImage": "._output_converter_image:OutputConversionImageToImage",
        "OutputConversionImageToTable": "._output_converter_image:OutputConversionImageToTable",
        "OutputConversionTable": "._output_converter_table:OutputConversionTable",
        "OutputConversionTimeSeries": "._output_converter_time_series:OutputConversionTimeSeries",
    },
)

__all__ = [
    "InputConversion",
    "InputConversionImage",
    "InputConversionTable",
    "InputConversionTimeSeries",
    "OutputConversion",
    "OutputConversionImageToColumn",
    "OutputConversionImageToImage",
    "OutputConversionImageToTable",
    "OutputConversionTable",
    "OutputConversionTimeSeries",
]



from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from safeds._utils import _structural_hash
from safeds._validation import _check_bounds, _ClosedBound

if TYPE_CHECKING:
    from safeds.data.image.containers import Image


class ImageSize:
    """
    A container for image size data.

    Parameters
    ----------
    width:
        the width of the image
    height:
        the height of the image
    channel:
        the channel of the image

    Raises
    ------
    OutOfBoundsError:
        if width or height are below 1
    ValueError
        if an invalid channel is given
    """

    def __init__(self, width: int, height: int, channel: int, *, _ignore_invalid_channel: bool = False) -> None:
        _check_bounds("width", width, lower_bound=_ClosedBound(1))
        _check_bounds("height", height, lower_bound=_ClosedBound(1))
        if not _ignore_invalid_channel and channel not in (1, 3, 4):
            raise ValueError(f"Channel {channel} is not a valid channel option. Use either 1, 3 or 4")
        _check_bounds("channel", channel, lower_bound=_ClosedBound(1))

        self._width = width
        self._height = height
        self._channel = channel

    @staticmethod
    def from_image(image: Image) -> ImageSize:
        """
        Create a `ImageSize` of a given image.

        Parameters
        ----------
        image:
            the given image for the `ImageSize`

        Returns
        -------
        image_size:
            the calculated `ImageSize`
        """
        return ImageSize(image.width, image.height, image.channel)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ImageSize):
            return NotImplemented
        return (self is other) or (
            self._width == other._width and self._height == other._height and self._channel == other._channel
        )

    def __hash__(self) -> int:
        return _structural_hash(self._width, self._height, self._channel)

    def __sizeof__(self) -> int:
        return sys.getsizeof(self._width) + sys.getsizeof(self._height) + sys.getsizeof(self._channel)

    def __str__(self) -> str:
        return f"{self._width}x{self._height}x{self._channel} (WxHxC)"

    @property
    def width(self) -> int:
        """
        Get the width of this `ImageSize` in pixels.

        Returns
        -------
        width:
            The width of this `ImageSize`.
        """
        return self._width

    @property
    def height(self) -> int:
        """
        Get the height of this `ImageSize` in pixels.

        Returns
        -------
        height:
            The height of this `ImageSize`.
        """
        return self._height

    @property
    def channel(self) -> int:
        """
        Get the channel of this `ImageSize` in pixels.

        Returns
        -------
        channel:
            The channel of this `ImageSize`.
        """
        return self._channel



from __future__ import annotations

import copy
from typing import TYPE_CHECKING, Generic, Self, TypeVar

from safeds._config import _init_default_device
from safeds._validation import _check_bounds, _ClosedBound
from safeds.data.image.containers import ImageList
from safeds.data.labeled.containers import ImageDataset, TabularDataset, TimeSeriesDataset
from safeds.data.tabular.containers import Table
from safeds.exceptions import (
    FeatureDataMismatchError,
    InputSizeError,
    InvalidModelStructureError,
    ModelNotFittedError,
)
from safeds.ml.nn.converters import (
    InputConversionImage,
    OutputConversionImageToColumn,
    OutputConversionImageToImage,
    OutputConversionImageToTable,
)
from safeds.ml.nn.converters._output_converter_image import _OutputConversionImage
from safeds.ml.nn.layers import (
    Convolutional2DLayer,
    FlattenLayer,
    ForwardLayer,
)
from safeds.ml.nn.layers._pooling2d_layer import _Pooling2DLayer

if TYPE_CHECKING:
    from collections.abc import Callable

    from torch import Tensor, nn

    from safeds.data.image.typing import ImageSize
    from safeds.ml.nn.converters import InputConversion, OutputConversion
    from safeds.ml.nn.layers import Layer


IFT = TypeVar("IFT", TabularDataset, TimeSeriesDataset, ImageDataset)  # InputFitType
IPT = TypeVar("IPT", Table, TimeSeriesDataset, ImageList)  # InputPredictType
OT = TypeVar("OT", TabularDataset, TimeSeriesDataset, ImageDataset)  # OutputType


class NeuralNetworkRegressor(Generic[IFT, IPT, OT]):
    """
    A NeuralNetworkRegressor is a neural network that is used for regression tasks.

    Parameters
    ----------
    input_conversion:
        to convert the input data for the neural network
    layers:
        a list of layers for the neural network to learn
    output_conversion:
        to convert the output data of the neural network back

    Raises
    ------
    InvalidModelStructureError
        if the defined model structure is invalid
    """

    def __init__(
        self,
        input_conversion: InputConversion[IFT, IPT],
        layers: list[Layer],
        output_conversion: OutputConversion[IPT, OT],
    ):
        if len(layers) == 0:
            raise InvalidModelStructureError("You need to provide at least one layer to a neural network.")
        if isinstance(input_conversion, InputConversionImage):
            if not isinstance(output_conversion, _OutputConversionImage):
                raise InvalidModelStructureError(
                    "The defined model uses an input conversion for images but no output conversion for images.",
                )
            elif isinstance(output_conversion, OutputConversionImageToColumn | OutputConversionImageToTable):
                raise InvalidModelStructureError(
                    "A NeuralNetworkRegressor cannot be used with images as input and 1-dimensional data as output.",
                )
            data_dimensions = 2
            for layer in layers:
                if data_dimensions == 2 and (isinstance(layer, Convolutional2DLayer | _Pooling2DLayer)):
                    continue
                elif data_dimensions == 2 and isinstance(layer, FlattenLayer):
                    data_dimensions = 1
                elif data_dimensions == 1 and isinstance(layer, ForwardLayer):
                    continue
                else:
                    raise InvalidModelStructureError(
                        (
                            "The 2-dimensional data has to be flattened before using a 1-dimensional layer."
                            if data_dimensions == 2
                            else "You cannot use a 2-dimensional layer with 1-dimensional data."
                        ),
                    )
            if data_dimensions == 1 and isinstance(output_conversion, OutputConversionImageToImage):
                raise InvalidModelStructureError(
                    "The output data would be 1-dimensional but the provided output conversion uses 2-dimensional data.",
                )
        elif isinstance(output_conversion, _OutputConversionImage):
            raise InvalidModelStructureError(
                "The defined model uses an output conversion for images but no input conversion for images.",
            )
        else:
            for layer in layers:
                if isinstance(layer, Convolutional2DLayer | FlattenLayer | _Pooling2DLayer):
                    raise InvalidModelStructureError("You cannot use a 2-dimensional layer with 1-dimensional data.")

        self._input_conversion: InputConversion[IFT, IPT] = input_conversion
        self._model = _create_internal_model(input_conversion, layers, is_for_classification=False)
        self._output_conversion: OutputConversion[IPT, OT] = output_conversion
        self._input_size = self._model.input_size
        self._batch_size = 1
        self._is_fitted = False
        self._total_number_of_batches_done = 0
        self._total_number_of_epochs_done = 0

    def fit(
        self,
        train_data: IFT,
        epoch_size: int = 25,
        batch_size: int = 1,
        learning_rate: float = 0.001,
        callback_on_batch_completion: Callable[[int, float], None] | None = None,
        callback_on_epoch_completion: Callable[[int, float], None] | None = None,
    ) -> Self:
        """
        Train the neural network with given training data.

        The original model is not modified.

        Parameters
        ----------
        train_data:
            The data the network should be trained on.
        epoch_size:
            The number of times the training cycle should be done.
        batch_size:
            The size of data batches that should be loaded at one time.
        learning_rate:
            The learning rate of the neural network.
        callback_on_batch_completion:
            Function used to view metrics while training. Gets called after a batch is completed with the index of the last batch and the overall loss average.
        callback_on_epoch_completion:
            Function used to view metrics while training. Gets called after an epoch is completed with the index of the last epoch and the overall loss average.

        Returns
        -------
        trained_model:
            The trained Model

        Raises
        ------
        OutOfBoundsError
            If epoch_size < 1
            If batch_size < 1
        """
        import torch
        from torch import nn

        _init_default_device()

        if not self._input_conversion._is_fit_data_valid(train_data):
            raise FeatureDataMismatchError

        _check_bounds("epoch_size", epoch_size, lower_bound=_ClosedBound(1))
        _check_bounds("batch_size", batch_size, lower_bound=_ClosedBound(1))

        if self._input_conversion._data_size is not self._input_size:
            raise InputSizeError(self._input_conversion._data_size, self._input_size)

        copied_model = copy.deepcopy(self)

        copied_model._batch_size = batch_size

        dataloader = copied_model._input_conversion._data_conversion_fit(train_data, copied_model._batch_size)

        loss_fn = nn.MSELoss()

        optimizer = torch.optim.SGD(copied_model._model.parameters(), lr=learning_rate)
        for _ in range(epoch_size):
            loss_sum = 0.0
            amount_of_loss_values_calculated = 0
            for x, y in iter(dataloader):
                optimizer.zero_grad()

                pred = copied_model._model(x)

                loss = loss_fn(pred, y)
                loss_sum += loss.item()
                amount_of_loss_values_calculated += 1
                loss.backward()
                optimizer.step()
                copied_model._total_number_of_batches_done += 1
                if callback_on_batch_completion is not None:
                    callback_on_batch_completion(
                        copied_model._total_number_of_batches_done,
                        loss_sum / amount_of_loss_values_calculated,
                    )
            copied_model._total_number_of_epochs_done += 1
            if callback_on_epoch_completion is not None:
                callback_on_epoch_completion(
                    copied_model._total_number_of_epochs_done,
                    loss_sum / amount_of_loss_values_calculated,
                )
        copied_model._is_fitted = True
        copied_model._model.eval()
        return copied_model

    def predict(self, test_data: IPT) -> OT:
        """
        Make a prediction for the given test data.

        The original Model is not modified.

        Parameters
        ----------
        test_data:
            The data the network should predict.

        Returns
        -------
        prediction:
            The given test_data with an added "prediction" column at the end

        Raises
        ------
        ModelNotFittedError
            If the model has not been fitted yet
        """
        import torch

        _init_default_device()

        if not self._is_fitted:
            raise ModelNotFittedError
        if not self._input_conversion._is_predict_data_valid(test_data):
            raise FeatureDataMismatchError
        dataloader = self._input_conversion._data_conversion_predict(test_data, self._batch_size)
        predictions = []
        with torch.no_grad():
            for x in dataloader:
                elem = self._model(x)
                predictions.append(elem.squeeze(dim=1))
        return self._output_conversion._data_conversion(
            test_data,
            torch.cat(predictions, dim=0),
            **self._input_conversion._get_output_configuration(),
        )

    @property
    def is_fitted(self) -> bool:
        """Whether the model is fitted."""
        return self._is_fitted


class NeuralNetworkClassifier(Generic[IFT, IPT, OT]):
    """
    A NeuralNetworkClassifier is a neural network that is used for classification tasks.

    Parameters
    ----------
    input_conversion:
        to convert the input data for the neural network
    layers:
        a list of layers for the neural network to learn
    output_conversion:
        to convert the output data of the neural network back

    Raises
    ------
    InvalidModelStructureError
        if the defined model structure is invalid
    """

    def __init__(
        self,
        input_conversion: InputConversion[IFT, IPT],
        layers: list[Layer],
        output_conversion: OutputConversion[IPT, OT],
    ):
        if len(layers) == 0:
            raise InvalidModelStructureError("You need to provide at least one layer to a neural network.")
        if isinstance(output_conversion, OutputConversionImageToImage):
            raise InvalidModelStructureError("A NeuralNetworkClassifier cannot be used with images as output.")
        elif isinstance(input_conversion, InputConversionImage):
            if not isinstance(output_conversion, _OutputConversionImage):
                raise InvalidModelStructureError(
                    "The defined model uses an input conversion for images but no output conversion for images.",
                )
            data_dimensions = 2
            for layer in layers:
                if data_dimensions == 2 and (isinstance(layer, Convolutional2DLayer | _Pooling2DLayer)):
                    continue
                elif data_dimensions == 2 and isinstance(layer, FlattenLayer):
                    data_dimensions = 1
                elif data_dimensions == 1 and isinstance(layer, ForwardLayer):
                    continue
                else:
                    raise InvalidModelStructureError(
                        (
                            "The 2-dimensional data has to be flattened before using a 1-dimensional layer."
                            if data_dimensions == 2
                            else "You cannot use a 2-dimensional layer with 1-dimensional data."
                        ),
                    )
            if data_dimensions == 2 and (
                isinstance(output_conversion, OutputConversionImageToColumn | OutputConversionImageToTable)
            ):
                raise InvalidModelStructureError(
                    "The output data would be 2-dimensional but the provided output conversion uses 1-dimensional data.",
                )
        elif isinstance(output_conversion, _OutputConversionImage):
            raise InvalidModelStructureError(
                "The defined model uses an output conversion for images but no input conversion for images.",
            )
        else:
            for layer in layers:
                if isinstance(layer, Convolutional2DLayer | FlattenLayer | _Pooling2DLayer):
                    raise InvalidModelStructureError("You cannot use a 2-dimensional layer with 1-dimensional data.")

        self._input_conversion: InputConversion[IFT, IPT] = input_conversion
        self._model = _create_internal_model(input_conversion, layers, is_for_classification=True)
        self._output_conversion: OutputConversion[IPT, OT] = output_conversion
        self._input_size = self._model.input_size
        self._batch_size = 1
        self._is_fitted = False
        self._num_of_classes = (
            layers[-1].output_size if isinstance(layers[-1].output_size, int) else -1
        )  # Is always int but linter doesn't know
        self._total_number_of_batches_done = 0
        self._total_number_of_epochs_done = 0

    def fit(
        self,
        train_data: IFT,
        epoch_size: int = 25,
        batch_size: int = 1,
        learning_rate: float = 0.001,
        callback_on_batch_completion: Callable[[int, float], None] | None = None,
        callback_on_epoch_completion: Callable[[int, float], None] | None = None,
    ) -> Self:
        """
        Train the neural network with given training data.

        The original model is not modified.

        Parameters
        ----------
        train_data:
            The data the network should be trained on.
        epoch_size:
            The number of times the training cycle should be done.
        batch_size:
            The size of data batches that should be loaded at one time.
        learning_rate:
            The learning rate of the neural network.
        callback_on_batch_completion:
            Function used to view metrics while training. Gets called after a batch is completed with the index of the last batch and the overall loss average.
        callback_on_epoch_completion:
            Function used to view metrics while training. Gets called after an epoch is completed with the index of the last epoch and the overall loss average.

        Returns
        -------
        trained_model:
            The trained Model

        Raises
        ------
        ValueError
            If epoch_size < 1
            If batch_size < 1
        """
        import torch
        from torch import nn

        _init_default_device()

        if not self._input_conversion._is_fit_data_valid(train_data):
            raise FeatureDataMismatchError

        _check_bounds("epoch_size", epoch_size, lower_bound=_ClosedBound(1))
        _check_bounds("batch_size", batch_size, lower_bound=_ClosedBound(1))

        if self._input_conversion._data_size is not self._input_size:
            raise InputSizeError(self._input_conversion._data_size, self._input_size)

        copied_model = copy.deepcopy(self)

        copied_model._batch_size = batch_size

        dataloader = copied_model._input_conversion._data_conversion_fit(
            train_data,
            copied_model._batch_size,
            copied_model._num_of_classes,
        )

        if copied_model._num_of_classes > 1:
            loss_fn = nn.CrossEntropyLoss()
        else:
            loss_fn = nn.BCELoss()

        optimizer = torch.optim.SGD(copied_model._model.parameters(), lr=learning_rate)
        for _ in range(epoch_size):
            loss_sum = 0.0
            amount_of_loss_values_calculated = 0
            for x, y in iter(dataloader):
                optimizer.zero_grad()
                pred = copied_model._model(x)

                loss = loss_fn(pred, y)
                loss_sum += loss.item()
                amount_of_loss_values_calculated += 1
                loss.backward()
                optimizer.step()

                copied_model._total_number_of_batches_done += 1
                if callback_on_batch_completion is not None:
                    callback_on_batch_completion(
                        copied_model._total_number_of_batches_done,
                        loss_sum / amount_of_loss_values_calculated,
                    )
            copied_model._total_number_of_epochs_done += 1
            if callback_on_epoch_completion is not None:
                callback_on_epoch_completion(
                    copied_model._total_number_of_epochs_done,
                    loss_sum / amount_of_loss_values_calculated,
                )
        copied_model._is_fitted = True
        copied_model._model.eval()
        return copied_model

    def predict(self, test_data: IPT) -> OT:
        """
        Make a prediction for the given test data.

        The original Model is not modified.

        Parameters
        ----------
        test_data:
            The data the network should predict.

        Returns
        -------
        prediction:
            The given test_data with an added "prediction" column at the end

        Raises
        ------
        ModelNotFittedError
            If the Model has not been fitted yet
        """
        import torch

        _init_default_device()

        if not self._is_fitted:
            raise ModelNotFittedError
        if not self._input_conversion._is_predict_data_valid(test_data):
            raise FeatureDataMismatchError
        dataloader = self._input_conversion._data_conversion_predict(test_data, self._batch_size)
        predictions = []
        with torch.no_grad():
            for x in dataloader:
                elem = self._model(x)
                if self._num_of_classes > 1:
                    predictions.append(torch.argmax(elem, dim=1))
                else:
                    predictions.append(elem.squeeze(dim=1).round())
        return self._output_conversion._data_conversion(
            test_data,
            torch.cat(predictions, dim=0),
            **self._input_conversion._get_output_configuration(),
        )

    @property
    def is_fitted(self) -> bool:
        """Whether the model is fitted."""
        return self._is_fitted


def _create_internal_model(
    input_conversion: InputConversion[IFT, IPT],
    layers: list[Layer],
    is_for_classification: bool,
) -> nn.Module:
    from torch import nn

    _init_default_device()

    class _InternalModel(nn.Module):
        def __init__(self, layers: list[Layer], is_for_classification: bool) -> None:
            super().__init__()
            self._layer_list = layers
            internal_layers = []
            previous_output_size = None

            for layer in layers:
                if previous_output_size is not None:
                    layer._set_input_size(previous_output_size)
                elif isinstance(input_conversion, InputConversionImage):
                    layer._set_input_size(input_conversion._data_size)
                if isinstance(layer, FlattenLayer | _Pooling2DLayer):
                    internal_layers.append(layer._get_internal_layer())
                else:
                    internal_layers.append(layer._get_internal_layer(activation_function="relu"))
                previous_output_size = layer.output_size

            if is_for_classification:
                internal_layers.pop()
                if isinstance(layers[-1].output_size, int) and layers[-1].output_size > 2:
                    internal_layers.append(layers[-1]._get_internal_layer(activation_function="none"))
                else:
                    internal_layers.append(layers[-1]._get_internal_layer(activation_function="sigmoid"))
            self._pytorch_layers = nn.Sequential(*internal_layers)

        @property
        def input_size(self) -> int | ImageSize:
            return self._layer_list[0].input_size

        def forward(self, x: Tensor) -> Tensor:
            for layer in self._pytorch_layers:
                x = layer(x)
            return x

    return _InternalModel(layers, is_for_classification)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from safeds.data.labeled.containers import TabularDataset
from safeds.data.tabular.containers import Column, Table

from ._output_converter import OutputConversion

if TYPE_CHECKING:
    from torch import Tensor


class OutputConversionTable(OutputConversion[Table, TabularDataset]):
    """The output conversion for a neural network, defines the output parameters for the neural network."""

    def __init__(self, prediction_name: str = "prediction") -> None:
        """
        Define the output parameters for the neural network in the output conversion.

        Parameters
        ----------
        prediction_name:
            The name of the new column where the prediction will be stored.
        """
        self._prediction_name = prediction_name

    def _data_conversion(self, input_data: Table, output_data: Tensor, **kwargs: Any) -> TabularDataset:  # noqa: ARG002
        return input_data.add_columns([Column(self._prediction_name, output_data.tolist())]).to_tabular_dataset(
            self._prediction_name,
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from safeds.data.labeled.containers import TabularDataset
from safeds.data.tabular.containers import Table

from ._input_converter import InputConversion

if TYPE_CHECKING:
    from torch.utils.data import DataLoader


class InputConversionTable(InputConversion[TabularDataset, Table]):
    """The input conversion for a neural network, defines the input parameters for the neural network."""

    def __init__(self) -> None:
        """Define the input parameters for the neural network in the input conversion."""
        self._target_name = ""
        self._time_name = ""
        self._feature_names: list[str] = []
        self._first = True

    @property
    def _data_size(self) -> int:
        return len(self._feature_names)

    def _data_conversion_fit(self, input_data: TabularDataset, batch_size: int, num_of_classes: int = 1) -> DataLoader:
        return input_data._into_dataloader_with_classes(
            batch_size,
            num_of_classes,
        )

    def _data_conversion_predict(self, input_data: Table, batch_size: int) -> DataLoader:
        return input_data._into_dataloader(batch_size)

    def _is_fit_data_valid(self, input_data: TabularDataset) -> bool:
        if self._first:
            self._feature_names = input_data.features.column_names
            self._target_name = input_data.target.name
            self._first = False
        return (sorted(input_data.features.column_names)).__eq__(sorted(self._feature_names))

    def _is_predict_data_valid(self, input_data: Table) -> bool:
        return (sorted(input_data.column_names)).__eq__(sorted(self._feature_names))

    def _get_output_configuration(self) -> dict[str, Any]:
        return {}



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from safeds._config import _init_default_device
from safeds._utils import _structural_hash
from safeds._validation import _check_bounds, _ClosedBound
from safeds.data.image.typing import ImageSize

from ._layer import Layer

if TYPE_CHECKING:
    from torch import Tensor, nn


def _create_internal_model(input_size: int, output_size: int, activation_function: str) -> nn.Module:
    from torch import nn

    _init_default_device()

    class _InternalLayer(nn.Module):
        def __init__(self, input_size: int, output_size: int, activation_function: str):
            super().__init__()
            self._layer = nn.Linear(input_size, output_size)
            match activation_function:
                case "sigmoid":
                    self._fn = nn.Sigmoid()
                case "relu":
                    self._fn = nn.ReLU()
                case "softmax":
                    self._fn = nn.Softmax()
                case "none":
                    self._fn = None
                case _:
                    raise ValueError("Unknown Activation Function: " + activation_function)

        def forward(self, x: Tensor) -> Tensor:
            return self._fn(self._layer(x)) if self._fn is not None else self._layer(x)

    return _InternalLayer(input_size, output_size, activation_function)


class ForwardLayer(Layer):
    def __init__(self, output_size: int, input_size: int | None = None):
        """
        Create a Feed Forward Layer.

        Parameters
        ----------
        input_size:
            The number of neurons in the previous layer
        output_size:
            The number of neurons in this layer

        Raises
        ------
        OutOfBoundsError
            If input_size < 1
            If output_size < 1

        """
        if input_size is not None:
            self._set_input_size(input_size=input_size)

        _check_bounds("output_size", output_size, lower_bound=_ClosedBound(1))

        self._output_size = output_size

    def _get_internal_layer(self, **kwargs: Any) -> nn.Module:
        if "activation_function" not in kwargs:
            raise ValueError(
                "The activation_function is not set. The internal layer can only be created when the activation_function is provided in the kwargs.",
            )
        else:
            activation_function: str = kwargs["activation_function"]
        return _create_internal_model(self._input_size, self._output_size, activation_function)

    @property
    def input_size(self) -> int:
        """
        Get the input_size of this layer.

        Returns
        -------
        result:
            The amount of values being passed into this layer.
        """
        return self._input_size

    @property
    def output_size(self) -> int:
        """
        Get the output_size of this layer.

        Returns
        -------
        result:
            The Number of Neurons in this layer.
        """
        return self._output_size

    def _set_input_size(self, input_size: int | ImageSize) -> None:
        if isinstance(input_size, ImageSize):
            raise TypeError("The input_size of a forward layer has to be of type int.")

        _check_bounds("input_size", input_size, lower_bound=_ClosedBound(1))

        self._input_size = input_size

    def __hash__(self) -> int:
        """
        Return a deterministic hash value for this forward layer.

        Returns
        -------
        hash:
            the hash value
        """
        return _structural_hash(self._input_size, self._output_size)

    def __eq__(self, other: object) -> bool:
        """
        Compare two forward layer instances.

        Returns
        -------
        equals:
            'True' if input and output size are equal, 'False' otherwise.
        """
        if not isinstance(other, ForwardLayer):
            return NotImplemented
        if self is other:
            return True
        return self._input_size == other._input_size and self._output_size == other._output_size

    def __sizeof__(self) -> int:
        """
        Return the complete size of this object.

        Returns
        -------
        size:
            Size of this object in bytes.
        """
        import sys

        return sys.getsizeof(self._input_size) + sys.getsizeof(self._output_size)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from safeds.data.image.containers import ImageList
from safeds.data.labeled.containers import ImageDataset, TabularDataset, TimeSeriesDataset
from safeds.data.tabular.containers import Table

if TYPE_CHECKING:
    from torch import Tensor

IT = TypeVar("IT", Table, TimeSeriesDataset, ImageList)
OT = TypeVar("OT", TabularDataset, TimeSeriesDataset, ImageDataset)


class OutputConversion(Generic[IT, OT], ABC):
    """The output conversion for a neural network, defines the output parameters for the neural network."""

    @abstractmethod
    def _data_conversion(self, input_data: IT, output_data: Tensor, **kwargs: Any) -> OT:
        pass  # pragma: no cover



from __future__ import annotations

import io
import sys
import warnings
from pathlib import Path
from typing import TYPE_CHECKING

from safeds._config import _get_device, _init_default_device
from safeds._utils import _structural_hash
from safeds.data.image._utils._image_transformation_error_and_warning_checks import (
    _check_add_noise_errors,
    _check_adjust_brightness_errors_and_warnings,
    _check_adjust_color_balance_errors_and_warnings,
    _check_adjust_contrast_errors_and_warnings,
    _check_blur_errors_and_warnings,
    _check_crop_errors_and_warnings,
    _check_resize_errors,
    _check_sharpen_errors_and_warnings,
)
from safeds.data.image.typing import ImageSize
from safeds.exceptions import IllegalFormatError

if TYPE_CHECKING:
    from numpy import dtype, ndarray
    from torch import Tensor


class Image:
    """
    A container for image data.

    Parameters
    ----------
    image_tensor:
        The image data as tensor.
    """

    _filter_edges_kernel_cache: Tensor | None = None

    @staticmethod
    def _filter_edges_kernel() -> Tensor:
        import torch

        _init_default_device()

        if Image._filter_edges_kernel_cache is None:
            Image._filter_edges_kernel_cache = (
                torch.tensor([[-1.0, -1.0, -1.0], [-1.0, 8.0, -1.0], [-1.0, -1.0, -1.0]])
                .unsqueeze(dim=0)
                .unsqueeze(dim=0)
                .to(_get_device())
            )
        if Image._filter_edges_kernel_cache.device != _get_device():
            Image._filter_edges_kernel_cache = Image._filter_edges_kernel_cache.to(_get_device())  # pragma: no cover

        return Image._filter_edges_kernel_cache

    @staticmethod
    def from_file(path: str | Path) -> Image:
        """
        Create an image from a file.

        Parameters
        ----------
        path:
            The path to the image file.

        Returns
        -------
        image:
            The image.

        Raises
        ------
        FileNotFoundError
            If the file of the path cannot be found
        """
        from torchvision.io import read_image

        if isinstance(path, str):
            path = Path(path)

        _init_default_device()

        if not path.is_file():
            raise FileNotFoundError(f"No such file or directory: '{path}'")

        return Image(image_tensor=read_image(str(path)).to(_get_device()))

    @staticmethod
    def from_bytes(data: bytes) -> Image:
        """
        Create an image from bytes.

        Parameters
        ----------
        data:
            The data of the image.

        Returns
        -------
        image:
            The image.
        """
        import torch
        import torchvision

        _init_default_device()

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message="The given buffer is not writable, and PyTorch does not support non-writable tensors.",
            )
            input_tensor = torch.frombuffer(data, dtype=torch.uint8)

        return Image(image_tensor=torchvision.io.decode_image(input_tensor).to(_get_device()))

    def __init__(self, image_tensor: Tensor) -> None:
        self._image_tensor: Tensor = image_tensor

    def __eq__(self, other: object) -> bool:
        """
        Compare two images.

        Parameters
        ----------
        other:
            The image to compare to.

        Returns
        -------
        equals:
            Whether the two images contain equal pixel data.
        """
        import torch

        _init_default_device()

        if not isinstance(other, Image):
            return NotImplemented
        return (self is other) or (
            self._image_tensor.size() == other._image_tensor.size()
            and torch.all(torch.eq(self._image_tensor, other._image_tensor)).item()
        )

    def __hash__(self) -> int:
        """
        Return a deterministic hash value for this image.

        Returns
        -------
        hash:
            The hash value.
        """
        return _structural_hash(self.width, self.height, self.channel)

    def __sizeof__(self) -> int:
        """
        Return the complete size of this object.

        Returns
        -------
        size:
            Size of this object in bytes.
        """
        return sys.getsizeof(self._image_tensor) + self._image_tensor.element_size() * self._image_tensor.nelement()

    def __array__(self, numpy_dtype: str | dtype | None = None) -> ndarray:
        """
        Return the image as a numpy array.

        Returns
        -------
        numpy_array:
            The image as numpy array.
        """
        from numpy import uint8

        return (
            self._image_tensor.permute(1, 2, 0)
            .detach()
            .cpu()
            .numpy()
            .astype(uint8 if numpy_dtype is None else numpy_dtype)
        )

    def _repr_jpeg_(self) -> bytes | None:
        """
        Return a JPEG image as bytes.

        If the image has an alpha channel return None.

        Returns
        -------
        jpeg:
            The image as JPEG.
        """
        import torch
        from torchvision.transforms.v2 import functional as func2
        from torchvision.utils import save_image

        _init_default_device()

        if self.channel == 4:
            return None
        buffer = io.BytesIO()
        if self.channel == 1:
            func2.to_pil_image(self._image_tensor, mode="L").save(buffer, format="jpeg")
        else:
            save_image(self._image_tensor.to(torch.float32) / 255, buffer, format="jpeg")
        buffer.seek(0)
        return buffer.read()

    def _repr_png_(self) -> bytes:
        """
        Return a PNG image as bytes.

        Returns
        -------
        png:
            The image as PNG.
        """
        import torch
        from torchvision.transforms.v2 import functional as func2
        from torchvision.utils import save_image

        _init_default_device()

        buffer = io.BytesIO()
        if self.channel == 1:
            func2.to_pil_image(self._image_tensor, mode="L").save(buffer, format="png")
        else:
            save_image(self._image_tensor.to(torch.float32) / 255, buffer, format="png")
        buffer.seek(0)
        return buffer.read()

    # ------------------------------------------------------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def width(self) -> int:
        """
        Get the width of the image in pixels.

        Returns
        -------
        width:
            The width of the image.
        """
        return self._image_tensor.size(dim=2)

    @property
    def height(self) -> int:
        """
        Get the height of the image in pixels.

        Returns
        -------
        height:
            The height of the image.
        """
        return self._image_tensor.size(dim=1)

    @property
    def channel(self) -> int:
        """
        Get the number of channels of the image.

        Returns
        -------
        channel:
            The number of channels of the image.
        """
        return self._image_tensor.size(dim=0)

    @property
    def size(self) -> ImageSize:
        """
        Get the `ImageSize` of the image.

        Returns
        -------
        image_size:
            The size of the image.
        """
        return ImageSize(self.width, self.height, self.channel)

    # ------------------------------------------------------------------------------------------------------------------
    # Conversion
    # ------------------------------------------------------------------------------------------------------------------

    def to_jpeg_file(self, path: str | Path) -> None:
        """
        Save the image as a JPEG file.

        Parameters
        ----------
        path:
            The path to the JPEG file.
        """
        import torch
        from torchvision.transforms.v2 import functional as func2
        from torchvision.utils import save_image

        _init_default_device()

        if self.channel == 4:
            raise IllegalFormatError("png")
        Path(path).parent.mkdir(parents=True, exist_ok=True)
        if self.channel == 1:
            func2.to_pil_image(self._image_tensor, mode="L").save(path, format="jpeg")
        else:
            save_image(self._image_tensor.to(torch.float32) / 255, path, format="jpeg")

    def to_png_file(self, path: str | Path) -> None:
        """
        Save the image as a PNG file.

        Parameters
        ----------
        path:
            The path to the PNG file.
        """
        import torch
        from torchvision.transforms.v2 import functional as func2
        from torchvision.utils import save_image

        _init_default_device()

        Path(path).parent.mkdir(parents=True, exist_ok=True)
        if self.channel == 1:
            func2.to_pil_image(self._image_tensor, mode="L").save(path, format="png")
        else:
            save_image(self._image_tensor.to(torch.float32) / 255, path, format="png")

    # ------------------------------------------------------------------------------------------------------------------
    # Transformations
    # ------------------------------------------------------------------------------------------------------------------

    def change_channel(self, channel: int) -> Image:
        """
        Return a new `Image` that has the given number of channels.

        The original image is not modified.

        Parameters
        ----------
        channel:
            The new number of channels. 1 will result in a grayscale image.

        Returns
        -------
        result:
            The image with the given number of channels.

        Raises
        ------
        ValueError
            if the given channel is not a valid channel option
        """
        import torch

        _init_default_device()

        if self.channel == channel:
            image_tensor = self._image_tensor
        elif self.channel == 1 and channel == 3:
            image_tensor = torch.cat([self._image_tensor, self._image_tensor, self._image_tensor], dim=0)
        elif self.channel == 1 and channel == 4:
            image_tensor = torch.cat(
                [
                    self._image_tensor,
                    self._image_tensor,
                    self._image_tensor,
                    torch.full(self._image_tensor.size(), 255),
                ],
                dim=0,
            )
        elif self.channel in (3, 4) and channel == 1:
            image_tensor = self.convert_to_grayscale()._image_tensor[0:1]
        elif self.channel == 3 and channel == 4:
            image_tensor = torch.cat(
                [self._image_tensor, torch.full(self._image_tensor[0:1].size(), 255)],
                dim=0,
            )
        elif self.channel == 4 and channel == 3:
            image_tensor = self._image_tensor[0:3]
        else:
            raise ValueError(f"Channel {channel} is not a valid channel option. Use either 1, 3 or 4")
        return Image(image_tensor)

    def resize(self, new_width: int, new_height: int) -> Image:
        """
        Return a new `Image` that has been resized to a given size.

        The original image is not modified.

        Parameters
        ----------
        new_width:
            the new width of the image
        new_height:
            the new height of the image

        Returns
        -------
        result:
            The image with the given width and height.

        Raises
        ------
        OutOfBoundsError
            If new_width or new_height are below 1
        """
        from torchvision.transforms import InterpolationMode
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        _check_resize_errors(new_width, new_height)
        return Image(
            func2.resize(self._image_tensor, size=[new_height, new_width], interpolation=InterpolationMode.NEAREST),
        )

    def convert_to_grayscale(self) -> Image:
        """
        Return a new `Image` that is converted to grayscale.

        The original image is not modified.

        Returns
        -------
        result:
            The grayscale image.
        """
        import torch
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        if self.channel == 4:
            return Image(
                torch.cat(
                    [
                        func2.rgb_to_grayscale(self._image_tensor[0:3], num_output_channels=3),
                        self._image_tensor[3].unsqueeze(dim=0),
                    ],
                ),
            )
        else:
            return Image(
                func2.rgb_to_grayscale(self._image_tensor[0:3], num_output_channels=self.channel),
            )

    def crop(self, x: int, y: int, width: int, height: int) -> Image:
        """
        Return a new `Image` that has been cropped to a given bounding rectangle.

        The original image is not modified.

        Parameters
        ----------
        x:
            the x coordinate of the top-left corner of the bounding rectangle
        y:
            the y coordinate of the top-left corner of the bounding rectangle
        width:
            the width of the bounding rectangle
        height:
            the height of the bounding rectangle

        Returns
        -------
        result:
            The cropped image.

        Raises
        ------
        OutOfBoundsError
            If x or y are below 0 or if width or height are below 1
        """
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        _check_crop_errors_and_warnings(x, y, width, height, self.width, self.height, plural=False)
        return Image(func2.crop(self._image_tensor, y, x, height, width))

    def flip_vertically(self) -> Image:
        """
        Return a new `Image` that is flipped vertically (horizontal axis, flips up-down and vice versa).

        The original image is not modified.

        Returns
        -------
        result:
            The flipped image.
        """
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        return Image(func2.vertical_flip(self._image_tensor))

    def flip_horizontally(self) -> Image:
        """
        Return a new `Image` that is flipped horizontally (vertical axis, flips left-right and vice versa).

        The original image is not modified.

        Returns
        -------
        result:
            The flipped image.
        """
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        return Image(func2.horizontal_flip(self._image_tensor))

    def adjust_brightness(self, factor: float) -> Image:
        """
        Return a new `Image` with an adjusted brightness.

        The original image is not modified.

        Parameters
        ----------
        factor:
            The brightness factor.
            1.0 will not change the brightness.
            Below 1.0 will result in a darker image.
            Above 1.0 will resolut in a brighter image.
            Has to be bigger than or equal to 0 (black).

        Returns
        -------
        result:
            The Image with adjusted brightness.

        Raises
        ------
        OutOfBoundsError
            If factor is smaller than 0.
        """
        import torch
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        _check_adjust_brightness_errors_and_warnings(factor, plural=False)
        if self.channel == 4:
            return Image(
                torch.cat(
                    [
                        func2.adjust_brightness(self._image_tensor[0:3], factor * 1.0),
                        self._image_tensor[3].unsqueeze(dim=0),
                    ],
                ),
            )
        else:
            return Image(func2.adjust_brightness(self._image_tensor, factor * 1.0))

    def add_noise(self, standard_deviation: float) -> Image:
        """
        Return a new `Image` with noise added to the image.

        The original image is not modified.

        Parameters
        ----------
        standard_deviation:
            The standard deviation of the normal distribution. Has to be bigger than or equal to 0.

        Returns
        -------
        result:
            The image with added noise.

        Raises
        ------
        OutOfBoundsError
            If standard_deviation is smaller than 0.
        """
        import torch

        _init_default_device()

        _check_add_noise_errors(standard_deviation)
        return Image(
            self._image_tensor + torch.normal(0, standard_deviation, self._image_tensor.size()).to(_get_device()) * 255,
        )

    def adjust_contrast(self, factor: float) -> Image:
        """
        Return a new `Image` with adjusted contrast.

        The original image is not modified.

        Parameters
        ----------
        factor:
            If factor > 1, increase contrast of image.
            If factor = 1, no changes will be made.
            If factor < 1, make image greyer.
            Has to be bigger than or equal to 0 (gray).

        Returns
        -------
        image:
            New image with adjusted contrast.

        Raises
        ------
        OutOfBoundsError
            If factor is smaller than 0.
        """
        import torch
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        _check_adjust_contrast_errors_and_warnings(factor, plural=False)
        if self.channel == 4:
            return Image(
                torch.cat(
                    [
                        func2.adjust_contrast(self._image_tensor[0:3], factor * 1.0),
                        self._image_tensor[3].unsqueeze(dim=0),
                    ],
                ),
            )
        else:
            return Image(func2.adjust_contrast(self._image_tensor, factor * 1.0))

    def adjust_color_balance(self, factor: float) -> Image:
        """
        Return a new `Image` with adjusted color balance.

        The original image is not modified.

        Parameters
        ----------
        factor:
            Has to be bigger than or equal to 0.
            If 0 <= factor < 1, make image greyer.
            If factor = 1, no changes will be made.
            If factor > 1, increase color balance of image.

        Returns
        -------
        image:
            The new, adjusted image.

        Raises
        ------
        OutOfBoundsError
            If factor is smaller than 0.
        """
        _check_adjust_color_balance_errors_and_warnings(factor, self.channel, plural=False)
        return Image(
            self.convert_to_grayscale()._image_tensor * (1.0 - factor * 1.0) + self._image_tensor * (factor * 1.0),
        )

    def blur(self, radius: int) -> Image:
        """
        Return a blurred version of the image.

        The original image is not modified.

        Parameters
        ----------
        radius:
             Radius is directly proportional to the blur value. The radius is equal to the amount of pixels united in
             each direction. A radius of 1 will result in a united box of 9 pixels.

        Returns
        -------
        result:
            The blurred image.

        Raises
        ------
        OutOfBoundsError
            If radius is smaller than 0 or equal or greater than the smaller size of the image.
        """
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        _check_blur_errors_and_warnings(radius, min(self.width, self.height), plural=False)
        return Image(func2.gaussian_blur(self._image_tensor, [radius * 2 + 1, radius * 2 + 1]))

    def sharpen(self, factor: float) -> Image:
        """
        Return a sharpened version of the image.

        The original image is not modified.

        Parameters
        ----------
        factor:
            If factor > 1, increase the sharpness of the image.
            If factor = 1, no changes will be made.
            If factor < 1, blur the image.
            Has to be bigger than or equal to 0 (blurred).

        Returns
        -------
        result:
            The image sharpened by the given factor.

        Raises
        ------
        OutOfBoundsError
            If factor is smaller than 0.
        """
        import torch
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        _check_sharpen_errors_and_warnings(factor, plural=False)
        if self.channel == 4:
            return Image(
                torch.cat(
                    [
                        func2.adjust_sharpness(self._image_tensor[0:3], factor * 1.0),
                        self._image_tensor[3].unsqueeze(dim=0),
                    ],
                ),
            )
        else:
            return Image(func2.adjust_sharpness(self._image_tensor, factor * 1.0))

    def invert_colors(self) -> Image:
        """
        Return a new `Image` with colors inverted.

        The original image is not modified.

        Returns
        -------
        result:
            The image with inverted colors.
        """
        import torch
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        if self.channel == 4:
            return Image(
                torch.cat([func2.invert(self._image_tensor[0:3]), self._image_tensor[3].unsqueeze(dim=0)]),
            )
        else:
            return Image(func2.invert(self._image_tensor))

    def rotate_right(self) -> Image:
        """
        Return a new `Image` that is rotated 90 degrees clockwise.

        The original image is not modified.

        Returns
        -------
        result:
            The image rotated 90 degrees clockwise.
        """
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        return Image(func2.rotate(self._image_tensor, -90, expand=True))

    def rotate_left(self) -> Image:
        """
        Return a new `Image` that is rotated 90 degrees counter-clockwise.

        The original image is not modified.

        Returns
        -------
        result:
            The image rotated 90 degrees counter-clockwise.
        """
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        return Image(func2.rotate(self._image_tensor, 90, expand=True))

    def find_edges(self) -> Image:
        """
        Return a grayscale version of the image with the edges highlighted.

        The original image is not modified.

        Returns
        -------
        result:
            The image with edges found.
        """
        import torch

        _init_default_device()

        edges_tensor = torch.clamp(
            torch.nn.functional.conv2d(
                self.convert_to_grayscale()._image_tensor.float()[0].unsqueeze(dim=0),
                Image._filter_edges_kernel(),
                padding="same",
            ).squeeze(dim=1),
            0,
            255,
        ).to(torch.uint8)
        if self.channel == 3:
            return Image(edges_tensor.repeat(3, 1, 1))
        elif self.channel == 4:
            return Image(
                torch.cat([edges_tensor.repeat(3, 1, 1), self._image_tensor[3].unsqueeze(dim=0)]),
            )
        else:
            return Image(edges_tensor)



from __future__ import annotations

from typing import TYPE_CHECKING

from safeds._utils import _structural_hash
from safeds.ml.classical._bases import _AdaBoostBase

from ._regressor import Regressor

if TYPE_CHECKING:
    from sklearn.base import RegressorMixin


class AdaBoostRegressor(Regressor, _AdaBoostBase):
    """
    Ada Boost regression.

    Parameters
    ----------
    learner:
        The learner from which the boosted ensemble is built.
    maximum_number_of_learners:
        The maximum number of learners at which boosting is terminated. In case of perfect fit, the learning procedure
        is stopped early. Has to be greater than 0.
    learning_rate:
        Weight applied to each regressor at each boosting iteration. A higher learning rate increases the contribution
        of each regressor. Has to be greater than 0.

    Raises
    ------
    OutOfBoundsError
        If `maximum_number_of_learners` or `learning_rate` are less than or equal to 0.
    """

    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(
        self,
        *,
        learner: Regressor | None = None,
        maximum_number_of_learners: int = 50,
        learning_rate: float = 1.0,
    ) -> None:
        # Initialize superclasses
        Regressor.__init__(self)
        _AdaBoostBase.__init__(
            self,
            maximum_number_of_learners=maximum_number_of_learners,
            learning_rate=learning_rate,
        )

        # Hyperparameters
        self._learner: Regressor | None = learner

    def __hash__(self) -> int:
        return _structural_hash(
            Regressor.__hash__(self),
            _AdaBoostBase.__hash__(self),
            self._learner,
        )

    # ------------------------------------------------------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def learner(self) -> Regressor | None:
        """The base learner used for training the ensemble."""
        return self._learner

    # ------------------------------------------------------------------------------------------------------------------
    # Template methods
    # ------------------------------------------------------------------------------------------------------------------

    def _clone(self) -> AdaBoostRegressor:
        return AdaBoostRegressor(
            learner=self.learner,
            maximum_number_of_learners=self._maximum_number_of_learners,
            learning_rate=self._learning_rate,
        )

    def _get_sklearn_model(self) -> RegressorMixin:
        from sklearn.ensemble import AdaBoostRegressor as SklearnAdaBoostRegressor

        learner = self.learner._get_sklearn_model() if self.learner is not None else None
        return SklearnAdaBoostRegressor(
            estimator=learner,
            n_estimators=self._maximum_number_of_learners,
            learning_rate=self._learning_rate,
        )



from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any

from safeds._utils import _structural_hash
from safeds.data.labeled.containers import TimeSeriesDataset
from safeds.data.tabular.containers import Column

from ._output_converter import OutputConversion

if TYPE_CHECKING:
    from torch import Tensor


class OutputConversionTimeSeries(OutputConversion[TimeSeriesDataset, TimeSeriesDataset]):
    """The output conversion for a neural network, defines the output parameters for the neural network."""

    def __hash__(self) -> int:
        """
        Return a deterministic hash value for this OutputConversionTimeSeries instance.

        Returns
        -------
        hash:
            the hash value
        """
        return _structural_hash(self.__class__.__name__ + self._prediction_name)

    def __eq__(self, other: object) -> bool:
        """
        Compare two OutputConversionTimeSeries instances.

        Parameters
        ----------
        other:
            The OutputConversionTimeSeries instance to compare to.

        Returns
        -------
        equals:
            Whether the instances are the same.
        """
        if not isinstance(other, OutputConversionTimeSeries):
            return False
        return self._prediction_name == other._prediction_name

    def __sizeof__(self) -> int:
        """
        Return the complete size of this object.

        Returns
        -------
        size:
            Size of this object in bytes.
        """
        return sys.getsizeof(self._prediction_name)

    def __init__(self, prediction_name: str = "prediction_nn") -> None:
        """
        Define the output parameters for the neural network in the output conversion.

        Parameters
        ----------
        prediction_name:
            The name of the new column where the prediction will be stored.
        """
        self._prediction_name = prediction_name

    def _data_conversion(self, input_data: TimeSeriesDataset, output_data: Tensor, **kwargs: Any) -> TimeSeriesDataset:
        if "window_size" not in kwargs or not isinstance(kwargs.get("window_size"), int):
            raise ValueError(
                "The window_size is not set. The data can only be converted if the window_size is provided as `int` in the kwargs.",
            )
        if "forecast_horizon" not in kwargs or not isinstance(kwargs.get("forecast_horizon"), int):
            raise ValueError(
                "The forecast_horizon is not set. The data can only be converted if the forecast_horizon is provided as `int` in the kwargs.",
            )
        window_size: int = kwargs["window_size"]
        forecast_horizon: int = kwargs["forecast_horizon"]
        input_data_table = input_data.to_table()
        input_data_table = input_data_table.slice_rows(start=window_size + forecast_horizon)

        return input_data_table.add_columns(
            [Column(self._prediction_name, output_data.tolist())],
        ).to_time_series_dataset(
            target_name=self._prediction_name,
            time_name=input_data.time.name,
            extra_names=input_data.extras.column_names,
        )



from __future__ import annotations

import io
import math
import os
import random
from abc import ABCMeta, abstractmethod
from pathlib import Path
from threading import Thread
from typing import TYPE_CHECKING, Literal, overload

from safeds._config import _init_default_device
from safeds._utils import _get_random_seed
from safeds._validation import _check_bounds, _ClosedBound
from safeds.data.image.containers._image import Image

if TYPE_CHECKING:
    from collections.abc import Sequence

    from torch import Tensor

    from safeds.data.image.containers._multi_size_image_list import _MultiSizeImageList
    from safeds.data.image.containers._single_size_image_list import _SingleSizeImageList
    from safeds.data.image.typing import ImageSize


class ImageList(metaclass=ABCMeta):
    """
    An ImageList is a list of different images. It can hold different sizes of Images. The channel of all images is the same.

    To create an `ImageList` call one of the following static methods:

    | Method                                                                        | Description                                              |
    | ----------------------------------------------------------------------------- | -------------------------------------------------------- |
    | [from_images][safeds.data.image.containers._image_list.ImageList.from_images] | Create an ImageList from a list of Images.               |
    | [from_files][safeds.data.image.containers._image_list.ImageList.from_files]   | Create an ImageList from a directory or a list of files. |
    """

    @staticmethod
    @abstractmethod
    def _create_image_list(images: list[Tensor], indices: list[int]) -> ImageList:
        """
        Create an ImageList from a list of tensors.

        Parameters
        ----------
        images:
            the list of tensors
        indices:
            a list of indices for the tensors

        Returns
        -------
        image_list:
            The image list
        """

    @staticmethod
    def from_images(images: list[Image]) -> ImageList:
        """
        Create an ImageList from a list of images.

        Parameters
        ----------
        images:
            the list of images

        Returns
        -------
        image_list:
            the image list
        """
        from safeds.data.image.containers._empty_image_list import _EmptyImageList
        from safeds.data.image.containers._multi_size_image_list import _MultiSizeImageList
        from safeds.data.image.containers._single_size_image_list import _SingleSizeImageList

        if len(images) == 0:
            return _EmptyImageList()

        indices = list(range(len(images)))
        first_width = images[0].width
        first_height = images[0].height
        for im in images:
            if first_width != im.width or first_height != im.height:
                return _MultiSizeImageList._create_image_list([image._image_tensor for image in images], indices)
        return _SingleSizeImageList._create_image_list([image._image_tensor for image in images], indices)

    @staticmethod
    @overload
    def from_files(path: str | Path | Sequence[str | Path]) -> ImageList: ...

    @staticmethod
    @overload
    def from_files(path: str | Path | Sequence[str | Path], *, load_percentage: float) -> ImageList: ...

    @staticmethod
    @overload
    def from_files(path: str | Path | Sequence[str | Path], *, return_filenames: Literal[False]) -> ImageList: ...

    @staticmethod
    @overload
    def from_files(
        path: str | Path | Sequence[str | Path],
        *,
        return_filenames: Literal[False],
        load_percentage: float,
    ) -> ImageList: ...

    @staticmethod
    @overload
    def from_files(
        path: str | Path | Sequence[str | Path],
        *,
        return_filenames: Literal[True],
    ) -> tuple[ImageList, list[str]]: ...

    @staticmethod
    @overload
    def from_files(
        path: str | Path | Sequence[str | Path],
        *,
        return_filenames: Literal[True],
        load_percentage: float,
    ) -> tuple[ImageList, list[str]]: ...

    @staticmethod
    @overload
    def from_files(
        path: str | Path | Sequence[str | Path],
        *,
        return_filenames: bool,
    ) -> ImageList | tuple[ImageList, list[str]]: ...

    @staticmethod
    @overload
    def from_files(
        path: str | Path | Sequence[str | Path],
        *,
        return_filenames: bool,
        load_percentage: float,
    ) -> ImageList | tuple[ImageList, list[str]]: ...

    @staticmethod
    def from_files(
        path: str | Path | Sequence[str | Path],
        *,
        return_filenames: bool = False,
        load_percentage: float = 1.0,
    ) -> ImageList | tuple[ImageList, list[str]]:
        """
        Create an ImageList from a directory or a list of files.

        If you provide a path to a directory the images will be sorted alphabetically while inner directories will be sorted after image files.

        Parameters
        ----------
        path:
            the path to the directory or a list of files
        return_filenames:
            if True the output will be a tuple which contains a list of the filenames in order of the images
        load_percentage:
            the percentage of the given data being loaded. If below 1 the files will be shuffled before loading

        Returns
        -------
        image_list:
            the image list

        Raises
        ------
        FileNotFoundError
            If the directory or one of the files of the path cannot be found
        OutOfBoundsError
            If load_percentage is not between 0 and 1
        """
        from PIL.Image import open as pil_image_open

        _init_default_device()

        random.seed(_get_random_seed())

        from safeds.data.image.containers._empty_image_list import _EmptyImageList
        from safeds.data.image.containers._multi_size_image_list import _MultiSizeImageList
        from safeds.data.image.containers._single_size_image_list import _SingleSizeImageList

        _check_bounds("load_percentage", load_percentage, lower_bound=_ClosedBound(0), upper_bound=_ClosedBound(1))

        if isinstance(path, list) and len(path) == 0:
            return _EmptyImageList()

        file_names = []

        path_list: list[str | Path]
        if isinstance(path, Path | str):
            path_list = [Path(path)]
        else:
            path_list = list(path)
        while len(path_list) != 0:
            p = Path(path_list.pop(0))
            if p.is_dir():
                path_list += sorted([p / name for name in os.listdir(p)])
            elif p.is_file():
                file_names.append(str(p))
            else:
                raise FileNotFoundError(f"No such file or directory: '{path}'")

        if load_percentage < 1:
            random.shuffle(file_names)
            file_names = file_names[: max(round(len(file_names) * load_percentage), 1) if load_percentage > 0 else 0]

        num_of_files = len(file_names)

        if num_of_files == 0:
            return _EmptyImageList()

        image_sizes: dict[tuple[int, int], dict[int, list[str]]] = {}
        image_indices: dict[tuple[int, int], dict[int, list[int]]] = {}
        image_count: dict[tuple[int, int], int] = {}
        max_channel = -1

        for i, filename in enumerate(file_names):
            im = pil_image_open(filename)
            im_channel = len(im.getbands())
            im_size = (im.width, im.height)
            if im_channel > max_channel:
                max_channel = im_channel
            if im_size not in image_sizes:
                image_sizes[im_size] = {im_channel: [filename]}
                image_indices[im_size] = {im_channel: [i]}
                image_count[im_size] = 1
            elif im_channel not in image_sizes[im_size]:
                image_sizes[im_size][im_channel] = [filename]
                image_indices[im_size][im_channel] = [i]
                image_count[im_size] += 1
            else:
                image_sizes[im_size][im_channel].append(filename)
                image_indices[im_size][im_channel].append(i)
                image_count[im_size] += 1

        num_of_threads = min(math.ceil(num_of_files / 1000), 100)
        num_of_files_per_thread = math.ceil(num_of_files / num_of_threads)

        single_sized_image_lists = []
        thread_packages = []
        for size, image_files in image_sizes.items():
            im_list, packages = _SingleSizeImageList._create_image_list_from_files(
                image_files,
                image_count[size],
                max_channel,
                size[0],
                size[1],
                image_indices[size],
                num_of_files_per_thread,
            )
            single_sized_image_lists.append(im_list._as_single_size_image_list())
            thread_packages += packages
        thread_packages.sort(key=lambda x: len(x), reverse=True)

        threads: list[ImageList._FromImageThread] = []
        for thread_index in range(num_of_threads):
            current_thread_workload = 0
            current_thread_packages = []
            while current_thread_workload < num_of_files_per_thread and len(thread_packages) > 0:
                next_package = thread_packages.pop()
                current_thread_packages.append(next_package)
                current_thread_workload += len(next_package)
            if thread_index == num_of_threads - 1 and len(thread_packages) > 0:
                current_thread_packages += thread_packages  # pragma: no cover
            thread = ImageList._FromImageThread(current_thread_packages)
            threads.append(thread)
            thread.start()
        for thread in threads:
            thread.join()

        if len(single_sized_image_lists) == 1:
            image_list: ImageList = single_sized_image_lists[0]
        else:
            image_list = _MultiSizeImageList._create_from_single_sized_image_lists(single_sized_image_lists)

        if return_filenames:
            return image_list, file_names
        else:
            return image_list

    class _FromFileThreadPackage:
        def __init__(
            self,
            im_files: list[str],
            im_channel: int,
            to_channel: int,
            im_width: int,
            im_height: int,
            tensor: Tensor,
            start_index: int,
        ) -> None:
            self._im_files = im_files
            self._im_channel = im_channel
            self._to_channel = to_channel
            self._im_width = im_width
            self._im_height = im_height
            self._tensor = tensor
            self._start_index = start_index

        def load_files(self) -> None:
            import torch
            from torchvision.io import read_image

            _init_default_device()

            num_of_files = len(self._im_files)
            tensor_channel = max(self._im_channel, min(self._to_channel, 3))
            for index, im in enumerate(self._im_files):
                self._tensor[index + self._start_index, 0:tensor_channel] = read_image(im)
            if self._to_channel == 4 and self._im_channel < 4:
                torch.full(
                    (num_of_files, 1, self._im_height, self._im_width),
                    255,
                    out=self._tensor[self._start_index : self._start_index + num_of_files, 3:4],
                )

        def __len__(self) -> int:
            return len(self._im_files)

    class _FromImageThread(Thread):
        def __init__(self, packages: list[ImageList._FromFileThreadPackage]) -> None:
            super().__init__()
            self._packages = packages

        def run(self) -> None:
            for pck in self._packages:
                pck.load_files()

    @abstractmethod
    def _clone(self) -> ImageList:
        """
        Clone your ImageList to a new instance.

        Returns
        -------
        image_list:
            the cloned image list
        """

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        """
        Compare two image lists.

        Two image lists are only equal if they contain the same images at the same indices.

        Parameters
        ----------
        other:
            the image list to compare to

        Returns
        -------
        equals:
            Whether the two image lists are equal
        """

    @abstractmethod
    def __hash__(self) -> int:
        """
        Return a deterministic hash value for this image list.

        Returns
        -------
        hash:
            The hash value.
        """

    @abstractmethod
    def __sizeof__(self) -> int:
        """
        Return the complete size of this object.

        Returns
        -------
        size:
            Size of this object in bytes.
        """

    def __len__(self) -> int:
        """
        Return the number of images in this image list.

        Returns
        -------
        length:
            The number of images
        """
        return self.number_of_images

    def __contains__(self, item: object) -> bool:
        """
        Return whether the given item is in this image list.

        Parameters
        ----------
        item:
            the item to check

        Returns
        -------
        has_item:
            Weather the given item is in this image list
        """
        return isinstance(item, Image) and self.has_image(item)

    def _repr_png_(self) -> bytes:
        """
        Return a PNG representation of this image list as bytes.

        Returns
        -------
        png:
            the png representation of this image list
        """
        import torch
        from torchvision.utils import make_grid, save_image

        _init_default_device()

        from safeds.data.image.containers._empty_image_list import _EmptyImageList

        if isinstance(self, _EmptyImageList):
            raise TypeError("You cannot display an empty ImageList")

        max_width, max_height = max(self.widths), max(self.heights)
        tensors = []
        for image in self.to_images():
            im_tensor = torch.zeros([4, max_height, max_width])
            im_tensor[:, : image.height, : image.width] = image.change_channel(4)._image_tensor
            tensors.append(im_tensor)
        tensor_grid = make_grid(tensors, math.ceil(math.sqrt(len(tensors))))
        buffer = io.BytesIO()
        save_image(tensor_grid.to(torch.float32) / 255, buffer, format="png")
        buffer.seek(0)
        return buffer.read()

    # ------------------------------------------------------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------------------------------------------------------

    @property
    @abstractmethod
    def number_of_images(self) -> int:
        """
        Return the number of images in this image list.

        Returns
        -------
        length:
            The number of images
        """

    @property
    @abstractmethod
    def widths(self) -> list[int]:
        """
        Return a list of all widths in this image list.

        Returns
        -------
        widths:
            A list of all widths
        """

    @property
    @abstractmethod
    def heights(self) -> list[int]:
        """
        Return a list of all heights in this image list.

        Returns
        -------
        heights:
            A list of all widths
        """

    @property
    @abstractmethod
    def channel(self) -> int:
        """
        Return the channel of all images.

        Returns
        -------
        channel:
            The channel of all images
        """

    @property
    @abstractmethod
    def sizes(self) -> list[ImageSize]:
        """
        Return the sizes of all images.

        Returns
        -------
        sizes:
            The sizes of all images
        """

    @property
    @abstractmethod
    def number_of_sizes(self) -> int:
        """
        Return the number of different sizes of images in this image list.

        Returns
        -------
        number_of_sizes:
            The number of different sizes
        """

    # ------------------------------------------------------------------------------------------------------------------
    # Getters
    # ------------------------------------------------------------------------------------------------------------------

    @abstractmethod
    def get_image(self, index: int) -> Image:
        """
        Return the image at the given index.

        Parameters
        ----------
        index:
            the index for the image to return

        Returns
        -------
        image:
            the image at the given index
        """

    @abstractmethod
    def index(self, image: Image) -> list[int]:
        """
        Return a list of indexes of the given image.

        If the image has multiple occurrences, all indices will be returned

        Parameters
        ----------
        image:
            the image to search for occurrences

        Returns
        -------
        indices:
            all occurrences of the image
        """

    @abstractmethod
    def has_image(self, image: Image) -> bool:
        """
        Return whether the given image is in this image list.

        Parameters
        ----------
        image:
            the image to check

        Returns
        -------
        has_image:
            Weather the given image is in this image list
        """

    # ------------------------------------------------------------------------------------------------------------------
    # Conversion
    # ------------------------------------------------------------------------------------------------------------------

    @abstractmethod
    def to_jpeg_files(self, path: str | Path | list[str | Path]) -> None:
        """
        Save all images as jpeg files.

        Parameters
        ----------
        path:
            Either the path to a directory or a list of directories which has directories for either all different sizes or all different images. Any non-existant path will be created

        Raises
        ------
        IllegalFormatError
            If the channel of the images is not supported
        ValueError
            If the path is a list but has too few or too many entries
        """

    @abstractmethod
    def to_png_files(self, path: str | Path | list[str | Path]) -> None:
        """
        Save all images as png files.

        Parameters
        ----------
        path:
            Either the path to a directory or a list of directories which has directories for either all different sizes or all different images. Any non-existant path will be created

        Raises
        ------
        ValueError
            If the path is a list but has too few or too many entries
        """

    @abstractmethod
    def to_images(self, indices: list[int] | None = None) -> list[Image]:
        """
        Return a list of all images in this image list.

        Parameters
        ----------
        indices:
            a list of all indices to include in the output. If None, all indices will be included

        Returns
        -------
        images:
            the list of all images

        Raises
        ------
        IndexOutOfBoundsError
            If any index is out of bounds
        """

    def _as_multi_size_image_list(self) -> _MultiSizeImageList:
        """
        Typechecking method for MultiSizeImageList.

        Returns
        -------
        self:
            self as a MultiSizeImageList

        Raises
        ------
        ValueError
            if this image list is not a MultiSizeImageList
        """
        from safeds.data.image.containers._multi_size_image_list import _MultiSizeImageList

        if isinstance(self, _MultiSizeImageList):
            return self
        raise ValueError("The given image_list is not a MultiSizeImageList")

    def _as_single_size_image_list(self) -> _SingleSizeImageList:
        """
        Typechecking method for SingleSizeImageList.

        Returns
        -------
        self:
            self as a SingleSizeImageList

        Raises
        ------
        ValueError
            if this image list is not a SingleSizeImageList
        """
        from safeds.data.image.containers._single_size_image_list import _SingleSizeImageList

        if isinstance(self, _SingleSizeImageList):
            return self
        raise ValueError("The given image_list is not a SingleSizeImageList")

    # ------------------------------------------------------------------------------------------------------------------
    # Transformations
    # ------------------------------------------------------------------------------------------------------------------

    @abstractmethod
    def change_channel(self, channel: int) -> ImageList:
        """
        Return a new `ImageList` that has the given number of channels.

        The original image list is not modified.

        Parameters
        ----------
        channel:
            The new number of channels. 1 will result in grayscale images

        Returns
        -------
        image_list:
            the image list with the given number of channels

        Raises
        ------
        ValueError:
            if the given channel is not a valid channel option
        """

    @abstractmethod
    def _add_image_tensor(self, image_tensor: Tensor, index: int) -> ImageList:
        """
        Return a new `ImageList` with the given tensor added as an image.

        The original image list is not modified.

        Parameters
        ----------
        image_tensor:
            The new tensor to be added to the image list

        Returns
        -------
        image_list:
            the image list with the new tensor added
        """

    def add_image(self, image: Image) -> ImageList:
        """
        Return a new `ImageList` with the given image added to the image list.

        The original image list is not modified.

        Parameters
        ----------
        image:
            The image to be added to the image list

        Returns
        -------
        image_list:
            the image list with the new image added
        """
        return self._add_image_tensor(image._image_tensor, self.number_of_images)

    @abstractmethod
    def add_images(self, images: list[Image] | ImageList) -> ImageList:
        """
        Return a new `ImageList` with the given images added to the image list.

        The original image list is not modified.

        Parameters
        ----------
        images:
            The images to be added to the image list

        Returns
        -------
        image_list:
            the image list with the new images added
        """

    def remove_image(self, image: Image) -> ImageList:
        """
        Return a new `ImageList` with the given image removed from the image list.

        If the image has multiple occurrences, all occurrences will be removed.

        The original image list is not modified.

        Parameters
        ----------
        image:
            The image to be removed from the image list

        Returns
        -------
        image_list:
            the image list with the given image removed
        """
        return self._remove_image_by_index_ignore_invalid(self.index(image))

    def remove_images(self, images: list[Image]) -> ImageList:
        """
        Return a new `ImageList` with the given images removed from the image list.

        If one image has multiple occurrences, all occurrences will be removed.

        The original image list is not modified.

        Parameters
        ----------
        images:
            The images to be removed from the image list

        Returns
        -------
        image_list:
            the image list with the given images removed
        """
        indices_to_remove = []
        for image in images:
            indices_to_remove += self.index(image)
        return self._remove_image_by_index_ignore_invalid(list(set(indices_to_remove)))

    @abstractmethod
    def remove_image_by_index(self, index: int | list[int]) -> ImageList:
        """
        Return a new `ImageList` with the given indices removed from the image list.

        The original image list is not modified.

        Parameters
        ----------
        index:
            The index of the image to be removed from the image list

        Returns
        -------
        image_list:
            the image list with the without the removed image

        Raises
        ------
        IndexOutOfBoundsError
            If one of the indices is out of bounds
        """

    @abstractmethod
    def _remove_image_by_index_ignore_invalid(self, index: int | list[int]) -> ImageList:
        """
        Return a new `ImageList` with the given indices removed from the image list.

        Invalid indices will be ignored.

        The original image list is not modified.

        Parameters
        ----------
        index:
            The index of the image to be removed from the image list

        Returns
        -------
        image_list:
            the image list with the without the removed image
        """

    @abstractmethod
    def remove_images_with_size(self, width: int, height: int) -> ImageList:
        """
        Return a new `ImageList` with the all images of the given size removed.

        The original image list is not modified.

        Parameters
        ----------
        width:
            The width of the images to be removed from the image list
        height:
            The height of the images to be removed from the image list

        Returns
        -------
        image_list:
            the image list with the given images removed

        Raises
        ------
        OutOfBoundsError
            If width or height are below 1
        """

    @abstractmethod
    def remove_duplicate_images(self) -> ImageList:
        """
        Return a new `ImageList` with all duplicate images removed.

        One occurrence of each image will stay in the image list.

        The original image list is not modified.

        Returns
        -------
        image_list:
            the image list with only unique images
        """

    @abstractmethod
    def shuffle_images(self) -> ImageList:
        """
        Return a new `ImageList` with all images shuffled.

        The original image list is not modified.

        Returns
        -------
        image_list:
            the image list with shuffled images
        """

    @abstractmethod
    def resize(self, new_width: int, new_height: int) -> ImageList:
        """
        Return a new `ImageList` with all images resized to a given size.

        The original image list is not modified.

        Parameters
        ----------
        new_width:
            the new width of the images
        new_height:
            the new height of the images

        Returns
        -------
        image_list:
            The image list with all images resized to the given width and height.

        Raises
        ------
        OutOfBoundsError
            If new_width or new_height are below 1
        """

    @abstractmethod
    def convert_to_grayscale(self) -> ImageList:
        """
        Return a new `ImageList` with all images converted to grayscale.

        The original image list is not modified.

        Returns
        -------
        image_list:
            The image list with all images converted to grayscale.
        """

    @abstractmethod
    def crop(self, x: int, y: int, width: int, height: int) -> ImageList:
        """
        Return a new `ImageList` with all images cropped to a given bounding rectangle.

        The original image list is not modified.

        Parameters
        ----------
        x:
            the x coordinate of the top-left corner of the bounding rectangle
        y:
            the y coordinate of the top-left corner of the bounding rectangle
        width:
            the width of the bounding rectangle
        height:
            the height of the bounding rectangle

        Returns
        -------
        image_list:
            The image list with all images cropped

        Raises
        ------
        OutOfBoundsError
            If x or y are below 0 or if width or height are below 1
        """

    @abstractmethod
    def flip_vertically(self) -> ImageList:
        """
        Return a new `ImageList` with all images flipped vertically (horizontal axis, flips up-down and vice versa).

        The original image list is not modified.

        Returns
        -------
        image_list:
            The image list with all images flipped vertically
        """

    @abstractmethod
    def flip_horizontally(self) -> ImageList:
        """
        Return a new `ImageList` with all images flipped horizontally (vertical axis, flips left-right and vice versa).

        The original image list is not modified.

        Returns
        -------
        image_list:
            The image list with all images flipped horizontally
        """

    @abstractmethod
    def adjust_brightness(self, factor: float) -> ImageList:
        """
        Return a new `ImageList` where all images have the adjusted brightness.

        The original image list is not modified.

        Parameters
        ----------
        factor:
            The brightness factor.
            1.0 will not change the brightness.
            Below 1.0 will result in a darker images.
            Above 1.0 will resolut in a brighter images.
            Has to be bigger than or equal to 0 (black).

        Returns
        -------
        image_list:
            The image list with adjusted brightness

        Raises
        ------
        OutOfBoundsError
            If factor is smaller than 0.
        """

    @abstractmethod
    def add_noise(self, standard_deviation: float) -> ImageList:
        """
        Return a new `ImageList` with noise added to all images.

        The original image list is not modified.

        Parameters
        ----------
        standard_deviation:
            The standard deviation of the normal distribution. Has to be bigger than or equal to 0.

        Returns
        -------
        image_list:
            The image list with added noise

        Raises
        ------
        OutOfBoundsError
            If standard_deviation is smaller than 0.
        """

    @abstractmethod
    def adjust_contrast(self, factor: float) -> ImageList:
        """
        Return a new `ImageList` where all images have the adjusted contrast.

        The original image list is not modified.

        Parameters
        ----------
        factor:
            If factor > 1, increase contrast of images.
            If factor = 1, no changes will be made.
            If factor < 1, make images greyer.
            Has to be bigger than or equal to 0 (gray).

        Returns
        -------
        image_list:
            The image list with adjusted contrast

        Raises
        ------
        OutOfBoundsError
            If factor is smaller than 0.
        """

    @abstractmethod
    def adjust_color_balance(self, factor: float) -> ImageList:
        """
        Return a new `ImageList` where all images have the adjusted color balance.

        The original image list is not modified.

        Parameters
        ----------
        factor:
            Has to be bigger than or equal to 0.
            If 0 <= factor < 1, make images greyer.
            If factor = 1, no changes will be made.
            If factor > 1, increase color balance of images.

        Returns
        -------
        image_list:
            The image list with adjusted color balance

        Raises
        ------
        OutOfBoundsError
            If factor is smaller than 0.
        """

    @abstractmethod
    def blur(self, radius: int) -> ImageList:
        """
        Return a new `ImageList` where all images have been blurred.

        The original image list is not modified.

        Parameters
        ----------
        radius:
             Radius is directly proportional to the blur value. The radius is equal to the amount of pixels united in
             each direction. A radius of 1 will result in a united box of 9 pixels.

        Returns
        -------
        image_list:
            The image list with blurred images

        Raises
        ------
        OutOfBoundsError
            If radius is smaller than 0 or equal or greater than the smallest size of one of the images.
        """

    @abstractmethod
    def sharpen(self, factor: float) -> ImageList:
        """
        Return a new `ImageList` where all images have been sharpened.

        The original image list is not modified.

        Parameters
        ----------
        factor:
            If factor > 1, increase the sharpness of the images.
            If factor = 1, no changes will be made.
            If factor < 1, blur the images.
            Has to be bigger than or equal to 0 (blurred).

        Returns
        -------
        image_list:
            The image list with sharpened images

        Raises
        ------
        OutOfBoundsError
            If factor is smaller than 0.
        """

    @abstractmethod
    def invert_colors(self) -> ImageList:
        """
        Return a new `ImageList` where all images have their colors inverted.

        The original image list is not modified.

        Returns
        -------
        image_list:
            The image list with inverted colors
        """

    @abstractmethod
    def rotate_right(self) -> ImageList:
        """
        Return a new `ImageList` where all images have been rotated 90 degrees clockwise.

        The original image list is not modified.

        Returns
        -------
        image_list:
            The image list with all images rotated
        """

    @abstractmethod
    def rotate_left(self) -> ImageList:
        """
        Return a new `ImageList` where all images have been rotated 90 degrees counter-clockwise.

        The original image list is not modified.

        Returns
        -------
        image_list:
            The image list with all images rotated
        """

    @abstractmethod
    def find_edges(self) -> ImageList:
        """
        Return a new `ImageList` with grayscale versions of the images with the edges highlighted.

        The original image list is not modified.

        Returns
        -------
        image_list:
            The image list with highlighted edges
        """



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from safeds._config import _init_default_device
from safeds._utils import _structural_hash
from safeds.data.image.containers import ImageList
from safeds.data.image.containers._single_size_image_list import _SingleSizeImageList
from safeds.data.labeled.containers import ImageDataset
from safeds.data.labeled.containers._image_dataset import _ColumnAsTensor, _TableAsTensor
from safeds.data.tabular.containers import Column, Table
from safeds.data.tabular.transformation import OneHotEncoder

from ._output_converter import OutputConversion

if TYPE_CHECKING:
    from torch import Tensor


class _OutputConversionImage(OutputConversion[ImageList, ImageDataset], ABC):
    @abstractmethod
    def _data_conversion(self, input_data: ImageList, output_data: Tensor, **kwargs: Any) -> ImageDataset:
        pass  # pragma: no cover

    def __hash__(self) -> int:
        """
        Return a deterministic hash value for this OutputConversionImage.

        Returns
        -------
        hash:
            the hash value
        """
        return _structural_hash(self.__class__.__name__)

    def __eq__(self, other: object) -> bool:
        """
        Compare two OutputConversionImage instances.

        Parameters
        ----------
        other:
            The OutputConversionImage instance to compare to.

        Returns
        -------
        equals:
            Whether the instances are the same.
        """
        if not isinstance(other, type(self)):
            return NotImplemented
        return True

    def __sizeof__(self) -> int:
        """
        Return the complete size of this object.

        Returns
        -------
        size:
            Size of this object in bytes.
        """
        return 0


class OutputConversionImageToColumn(_OutputConversionImage):
    def _data_conversion(self, input_data: ImageList, output_data: Tensor, **kwargs: Any) -> ImageDataset[Column]:
        import torch

        _init_default_device()

        if not isinstance(input_data, _SingleSizeImageList):
            raise ValueError("The given input ImageList contains images of different sizes.")  # noqa: TRY004
        if "column_name" not in kwargs or not isinstance(kwargs.get("column_name"), str):
            raise ValueError(
                "The column_name is not set. The data can only be converted if the column_name is provided as `str` in the kwargs.",
            )
        if "one_hot_encoder" not in kwargs or not isinstance(kwargs.get("one_hot_encoder"), OneHotEncoder):
            raise ValueError(
                "The one_hot_encoder is not set. The data can only be converted if the one_hot_encoder is provided as `OneHotEncoder` in the kwargs.",
            )
        one_hot_encoder: OneHotEncoder = kwargs["one_hot_encoder"]
        column_name: str = kwargs["column_name"]

        output = torch.zeros(len(input_data), len(one_hot_encoder._get_names_of_added_columns()))
        output[torch.arange(len(input_data)), output_data] = 1

        im_dataset: ImageDataset[Column] = ImageDataset[Column].__new__(ImageDataset)
        im_dataset._output = _ColumnAsTensor._from_tensor(output, column_name, one_hot_encoder)
        im_dataset._shuffle_tensor_indices = torch.LongTensor(list(range(len(input_data))))
        im_dataset._shuffle_after_epoch = False
        im_dataset._batch_size = 1
        im_dataset._next_batch_index = 0
        im_dataset._input_size = input_data.sizes[0]
        im_dataset._input = input_data
        return im_dataset


class OutputConversionImageToTable(_OutputConversionImage):
    def _data_conversion(self, input_data: ImageList, output_data: Tensor, **kwargs: Any) -> ImageDataset[Table]:
        import torch

        _init_default_device()

        if not isinstance(input_data, _SingleSizeImageList):
            raise ValueError("The given input ImageList contains images of different sizes.")  # noqa: TRY004
        if (
            "column_names" not in kwargs
            or not isinstance(kwargs.get("column_names"), list)
            and all(isinstance(element, str) for element in kwargs["column_names"])
        ):
            raise ValueError(
                "The column_names are not set. The data can only be converted if the column_names are provided as `list[str]` in the kwargs.",
            )
        column_names: list[str] = kwargs["column_names"]

        output = torch.zeros(len(input_data), len(column_names))
        output[torch.arange(len(input_data)), output_data] = 1

        im_dataset: ImageDataset[Table] = ImageDataset[Table].__new__(ImageDataset)
        im_dataset._output = _TableAsTensor._from_tensor(output, column_names)
        im_dataset._shuffle_tensor_indices = torch.LongTensor(list(range(len(input_data))))
        im_dataset._shuffle_after_epoch = False
        im_dataset._batch_size = 1
        im_dataset._next_batch_index = 0
        im_dataset._input_size = input_data.sizes[0]
        im_dataset._input = input_data
        return im_dataset


class OutputConversionImageToImage(_OutputConversionImage):
    def _data_conversion(
        self,
        input_data: ImageList,
        output_data: Tensor,
        **kwargs: Any,  # noqa: ARG002
    ) -> ImageDataset[ImageList]:
        import torch

        _init_default_device()

        if not isinstance(input_data, _SingleSizeImageList):
            raise ValueError("The given input ImageList contains images of different sizes.")  # noqa: TRY004

        return ImageDataset[ImageList](
            input_data,
            _SingleSizeImageList._create_from_tensor(
                (output_data * 255).to(torch.uint8),
                list(range(output_data.size(dim=0))),
            ),
        )



from __future__ import annotations

import copy
import random
import sys
from pathlib import Path
from typing import TYPE_CHECKING

from safeds._config import _get_device, _init_default_device
from safeds._utils import _structural_hash
from safeds.data.image._utils._image_transformation_error_and_warning_checks import (
    _check_add_noise_errors,
    _check_adjust_brightness_errors_and_warnings,
    _check_adjust_color_balance_errors_and_warnings,
    _check_adjust_contrast_errors_and_warnings,
    _check_blur_errors_and_warnings,
    _check_crop_errors_and_warnings,
    _check_remove_images_with_size_errors,
    _check_resize_errors,
    _check_sharpen_errors_and_warnings,
)
from safeds.data.image.containers._image import Image
from safeds.data.image.containers._image_list import ImageList
from safeds.data.image.typing import ImageSize
from safeds.exceptions import (
    DuplicateIndexError,
    IllegalFormatError,
    IndexOutOfBoundsError,
)

if TYPE_CHECKING:
    from torch import Tensor

    from safeds.data.image.containers._multi_size_image_list import _MultiSizeImageList


class _SingleSizeImageList(ImageList):
    """
    An ImageList is a list of different images. It can hold different sizes of Images. The channel of all images is the same.

    This is the class for an ImageList with only one size for all images.

    To create an `ImageList` call one of the following static methods:

    | Method                                                                        | Description                                              |
    | ----------------------------------------------------------------------------- | -------------------------------------------------------- |
    | [from_images][safeds.data.image.containers._image_list.ImageList.from_images] | Create an ImageList from a list of Images.               |
    | [from_files][safeds.data.image.containers._image_list.ImageList.from_files]   | Create an ImageList from a directory or a list of files. |
    """

    def __init__(self) -> None:
        import torch

        _init_default_device()

        self._next_batch_index = 0
        self._batch_size = 1

        self._tensor: Tensor = torch.empty([])
        self._tensor_positions_to_indices: list[int] = []  # list[tensor_position] = index
        self._indices_to_tensor_positions: dict[int, int] = {}  # {index: tensor_position}

    @staticmethod
    def _create_image_list_from_files(
        images: dict[int, list[str]],
        number_of_images: int,
        max_channel: int,
        width: int,
        height: int,
        indices: dict[int, list[int]],
        max_files_per_thread_package: int,
    ) -> tuple[ImageList, list[ImageList._FromFileThreadPackage]]:
        import torch

        _init_default_device()

        from safeds.data.image.containers._empty_image_list import _EmptyImageList

        if len(images) == 0 or number_of_images == 0:
            return _EmptyImageList(), []

        image_list = _SingleSizeImageList()

        images_tensor = torch.empty(
            number_of_images,
            max_channel,
            height,
            width,
            dtype=torch.uint8,
            device=_get_device(),
        )

        thread_packages: list[ImageList._FromFileThreadPackage] = []
        current_thread_channel: int | None = None
        current_thread_channel_files: list[str] = []
        current_thread_start_index: int = 0
        while number_of_images - current_thread_start_index > 0:
            num_of_files = min(max_files_per_thread_package, number_of_images - current_thread_start_index)
            while num_of_files > 0:
                if current_thread_channel is None or len(current_thread_channel_files) == 0:
                    current_thread_channel = next(iter(images.keys()))
                    current_thread_channel_files = images.pop(current_thread_channel)
                next_package_size = min(num_of_files, len(current_thread_channel_files))
                package = ImageList._FromFileThreadPackage(
                    current_thread_channel_files[:next_package_size],
                    current_thread_channel,
                    max_channel,
                    width,
                    height,
                    images_tensor,
                    current_thread_start_index,
                )
                current_thread_start_index += next_package_size
                num_of_files -= next_package_size
                current_thread_channel_files = current_thread_channel_files[next_package_size:]
                thread_packages.append(package)

        image_list._tensor = images_tensor
        image_list._tensor_positions_to_indices = [i for j in indices.values() for i in j]
        image_list._indices_to_tensor_positions = image_list._calc_new_indices_to_tensor_positions()
        return image_list, thread_packages

    @staticmethod
    def _create_image_list(images: list[Tensor], indices: list[int]) -> ImageList:
        import torch

        _init_default_device()

        from safeds.data.image.containers._empty_image_list import _EmptyImageList

        if len(images) == 0:
            return _EmptyImageList()

        image_list = _SingleSizeImageList()
        images_ready_to_concat: list[tuple[Tensor, int]] = []
        images_with_correct_channel: list[tuple[Tensor, int]] = []
        images_with_less_channels: list[tuple[Tensor, int]] = []
        max_channel = 0
        for image, index in zip(images, indices, strict=False):
            if max_channel < image.size(dim=-3):  # all images have to increase their channel
                images_with_less_channels += images_with_correct_channel
                images_with_correct_channel = [(image, index)]
                max_channel = image.size(dim=-3)
            elif max_channel > image.size(dim=-3):  # current image has to increase its channel
                images_with_less_channels.append((image, index))
            else:  # current image has same channel as max_channel
                images_with_correct_channel.append((image, index))
        for image, index in images_with_correct_channel:
            images_ready_to_concat.append((image.unsqueeze(dim=0), index))
        for image, index in images_with_less_channels:
            if max_channel == 3:  # image channel 1 and max channel 3
                image_to_append = torch.cat([image, image, image], dim=0)
            elif image.size(dim=0) == 1:  # image channel 1 and max channel 4
                image_to_append = torch.cat(
                    [image, image, image, torch.full(image.size(), 255, device=image.device)],
                    dim=0,
                )
            else:  # image channel 3 and max channel 4
                image_to_append = torch.cat([image, torch.full(image[0:1].size(), 255, device=image.device)], dim=0)
            images_ready_to_concat.append((image_to_append.unsqueeze(dim=0), index))
        image_list._tensor = torch.cat([image for image, index in images_ready_to_concat])
        image_list._tensor_positions_to_indices = [index for image, index in images_ready_to_concat]
        image_list._indices_to_tensor_positions = image_list._calc_new_indices_to_tensor_positions()

        return image_list

    @staticmethod
    def _create_from_tensor(images_tensor: Tensor, indices: list[int]) -> _SingleSizeImageList:
        if images_tensor.dim() == 3:
            images_tensor = images_tensor.unsqueeze(dim=1)
        if images_tensor.dim() != 4:
            raise ValueError(f"Invalid Tensor. This Tensor requires 3 or 4 dimensions but has {images_tensor.dim()}")

        image_list = _SingleSizeImageList()
        image_list._tensor = images_tensor.detach().clone()
        image_list._tensor_positions_to_indices = indices
        image_list._indices_to_tensor_positions = image_list._calc_new_indices_to_tensor_positions()

        return image_list

    def __iter__(self) -> _SingleSizeImageList:
        im_ds = copy.copy(self)
        im_ds._next_batch_index = 0
        return im_ds

    def __next__(self) -> Tensor:
        if self._next_batch_index * self._batch_size >= len(self):
            raise StopIteration
        self._next_batch_index += 1
        return self._get_batch(self._next_batch_index - 1)

    def _get_batch(self, batch_number: int, batch_size: int | None = None) -> Tensor:
        import torch

        _init_default_device()

        if batch_size is None:
            batch_size = self._batch_size
        if batch_size * batch_number >= len(self):
            raise IndexOutOfBoundsError(batch_size * batch_number)
        max_index = batch_size * (batch_number + 1) if batch_size * (batch_number + 1) < len(self) else len(self)
        return (
            self._tensor[
                [self._indices_to_tensor_positions[index] for index in range(batch_size * batch_number, max_index)]
            ].to(torch.float32)
            / 255
        )

    def _clone(self) -> ImageList:
        cloned_image_list = self._clone_without_tensor()
        cloned_image_list._tensor = self._tensor.detach().clone()
        return cloned_image_list

    def _clone_without_tensor(self) -> _SingleSizeImageList:
        """
        Clone this SingleSizeImageList to a new instance without the image data.

        Returns
        -------
        image_list:
            the cloned image list
        """
        cloned_image_list = _SingleSizeImageList()
        cloned_image_list._indices_to_tensor_positions = copy.deepcopy(self._indices_to_tensor_positions)
        cloned_image_list._tensor_positions_to_indices = copy.deepcopy(self._tensor_positions_to_indices)
        return cloned_image_list

    def _calc_new_indices_to_tensor_positions(self) -> dict[int, int]:
        """
        Calculate the new indices to tensor position dictionary.

        Returns
        -------
        new_indices_to_tensor_positions:
            the new dictionary
        """
        _indices_to_tensor_positions: dict[int, int] = {}
        for i, index in enumerate(self._tensor_positions_to_indices):
            _indices_to_tensor_positions[index] = i
        return _indices_to_tensor_positions

    def __eq__(self, other: object) -> bool:
        import torch

        _init_default_device()

        if not isinstance(other, ImageList):
            return NotImplemented
        if not isinstance(other, _SingleSizeImageList):
            return False
        return (self is other) or (
            self._tensor.size() == other._tensor.size()
            and set(self._tensor_positions_to_indices) == set(self._tensor_positions_to_indices)
            and set(self._indices_to_tensor_positions) == set(self._indices_to_tensor_positions)
            and torch.all(
                torch.eq(
                    self._tensor[torch.tensor(self._tensor_positions_to_indices).argsort()],
                    other._tensor[torch.tensor(other._tensor_positions_to_indices).argsort()],
                ),
            ).item()
        )

    def __hash__(self) -> int:
        return _structural_hash(
            self.widths[0],
            self.heights[0],
            self.channel,
            self.number_of_images,
            self._tensor_positions_to_indices,
        )

    def __sizeof__(self) -> int:
        return (
            sys.getsizeof(self._tensor)
            + self._tensor.element_size() * self._tensor.nelement()
            + sum(map(sys.getsizeof, self._tensor_positions_to_indices))
            + sys.getsizeof(self._tensor_positions_to_indices)
            + sum(map(sys.getsizeof, self._indices_to_tensor_positions.keys()))
            + sum(map(sys.getsizeof, self._indices_to_tensor_positions.values()))
            + sys.getsizeof(self._indices_to_tensor_positions)
        )

    @property
    def number_of_images(self) -> int:
        return self._tensor.size(dim=0)

    @property
    def widths(self) -> list[int]:
        return [self._tensor.size(dim=3)] * self.number_of_images

    @property
    def heights(self) -> list[int]:
        return [self._tensor.size(dim=2)] * self.number_of_images

    @property
    def channel(self) -> int:
        return self._tensor.size(dim=1)

    @property
    def sizes(self) -> list[ImageSize]:
        return [
            ImageSize(self._tensor.size(dim=3), self._tensor.size(dim=2), self._tensor.size(dim=1)),
        ] * self.number_of_images

    @property
    def number_of_sizes(self) -> int:
        return 1

    def get_image(self, index: int) -> Image:
        if index not in self._indices_to_tensor_positions:
            raise IndexOutOfBoundsError(index)
        return Image(self._tensor[self._indices_to_tensor_positions[index]])

    def index(self, image: Image) -> list[int]:
        if image not in self:
            return []
        return [
            self._tensor_positions_to_indices[i]
            for i in (image._image_tensor == self._tensor).all(dim=[1, 2, 3]).nonzero()
        ]

    def has_image(self, image: Image) -> bool:
        return (
            not (image.width != self.widths[0] or image.height != self.heights[0] or image.channel != self.channel)
            and image._image_tensor in self._tensor
        )

    def to_jpeg_files(self, path: str | Path | list[str | Path]) -> None:
        import torch
        from torchvision.transforms.v2 import functional as func2
        from torchvision.utils import save_image

        _init_default_device()

        if self.channel == 4:
            raise IllegalFormatError("png")
        path_str: str | Path
        if isinstance(path, list):
            if len(path) == self.number_of_images:
                for image_path, index in zip(path, sorted(self._tensor_positions_to_indices), strict=False):
                    Path(image_path).parent.mkdir(parents=True, exist_ok=True)
                    if self.channel == 1:
                        func2.to_pil_image(self._tensor[self._indices_to_tensor_positions[index]], mode="L").save(
                            image_path,
                            format="jpeg",
                        )
                    else:
                        save_image(
                            self._tensor[self._indices_to_tensor_positions[index]].to(torch.float32) / 255,
                            image_path,
                            format="jpeg",
                        )
                return
            elif len(path) == 1:
                path_str = path[0]
            else:
                raise ValueError(
                    "The path specified is invalid. Please provide either the path to a directory, a list of paths with one path for each image, or a list of paths with one path per image size.",
                )
        else:
            path_str = path
        for index in self._tensor_positions_to_indices:
            image_path = Path(path_str) / (str(index) + ".jpg")
            Path(image_path).parent.mkdir(parents=True, exist_ok=True)
            if self.channel == 1:
                func2.to_pil_image(self._tensor[self._indices_to_tensor_positions[index]], mode="L").save(
                    image_path,
                    format="jpeg",
                )
            else:
                save_image(
                    self._tensor[self._indices_to_tensor_positions[index]].to(torch.float32) / 255,
                    image_path,
                    format="jpeg",
                )

    def to_png_files(self, path: str | Path | list[str | Path]) -> None:
        import torch
        from torchvision.transforms.v2 import functional as func2
        from torchvision.utils import save_image

        _init_default_device()

        path_str: str | Path
        if isinstance(path, list):
            if len(path) == self.number_of_images:
                for image_path, index in zip(path, sorted(self._tensor_positions_to_indices), strict=False):
                    Path(image_path).parent.mkdir(parents=True, exist_ok=True)
                    if self.channel == 1:
                        func2.to_pil_image(self._tensor[self._indices_to_tensor_positions[index]], mode="L").save(
                            image_path,
                            format="png",
                        )
                    else:
                        save_image(
                            self._tensor[self._indices_to_tensor_positions[index]].to(torch.float32) / 255,
                            image_path,
                            format="png",
                        )
                return
            elif len(path) == 1:
                path_str = path[0]
            else:
                raise ValueError(
                    "The path specified is invalid. Please provide either the path to a directory, a list of paths with one path for each image, or a list of paths with one path per image size.",
                )
        else:
            path_str = path
        for index in self._tensor_positions_to_indices:
            image_path = Path(path_str) / (str(index) + ".png")
            Path(image_path).parent.mkdir(parents=True, exist_ok=True)
            if self.channel == 1:
                func2.to_pil_image(self._tensor[self._indices_to_tensor_positions[index]], mode="L").save(
                    image_path,
                    format="png",
                )
            else:
                save_image(
                    self._tensor[self._indices_to_tensor_positions[index]].to(torch.float32) / 255,
                    image_path,
                    format="png",
                )

    def to_images(self, indices: list[int] | None = None) -> list[Image]:
        if indices is None:
            indices = sorted(self._tensor_positions_to_indices)
        else:
            wrong_indices = []
            for index in indices:
                if index not in self._indices_to_tensor_positions:
                    wrong_indices.append(index)
            if len(wrong_indices) != 0:
                raise IndexOutOfBoundsError(wrong_indices)
        return [Image(self._tensor[self._indices_to_tensor_positions[index]]) for index in indices]

    def change_channel(self, channel: int) -> ImageList:
        if channel == self.channel:
            return self
        image_list = self._clone_without_tensor()
        image_list._tensor = _SingleSizeImageList._change_channel_of_tensor(self._tensor, channel)
        return image_list

    @staticmethod
    def _change_channel_of_tensor(tensor: Tensor, channel: int) -> Tensor:
        """
        Change the channel of a tensor to the given channel.

        Parameters
        ----------
        tensor:
            the tensor to change the channel of
        channel:
            the given new channel

        Returns
        -------
        new_tensor:
            a new tensor with the given channel

        Raises
        ------
        ValueError
            if the given channel is not a valid channel option
        """
        import torch

        _init_default_device()

        if tensor.size(dim=-3) == channel:
            return tensor.detach().clone()
        elif tensor.size(dim=-3) == 1 and channel == 3:
            return torch.cat([tensor, tensor, tensor], dim=1)
        elif tensor.size(dim=-3) == 1 and channel == 4:
            return torch.cat([tensor, tensor, tensor, torch.full(tensor.size(), 255)], dim=1)
        elif tensor.size(dim=-3) in (3, 4) and channel == 1:
            return _SingleSizeImageList._convert_tensor_to_grayscale(tensor)[:, 0:1]
        elif tensor.size(dim=-3) == 3 and channel == 4:
            return torch.cat([tensor, torch.full(tensor[:, 0:1].size(), 255)], dim=1)
        elif tensor.size(dim=-3) == 4 and channel == 3:
            return tensor[:, 0:3].detach().clone()
        else:
            raise ValueError(f"Channel {channel} is not a valid channel option. Use either 1, 3 or 4")

    def _add_image_tensor(self, image_tensor: Tensor, index: int) -> ImageList:
        import torch

        _init_default_device()

        from safeds.data.image.containers._multi_size_image_list import _MultiSizeImageList

        if index in self._indices_to_tensor_positions:
            raise DuplicateIndexError(index)

        if self._tensor.size(dim=2) == image_tensor.size(dim=1) and self._tensor.size(dim=3) == image_tensor.size(
            dim=2,
        ):
            image_list_single: _SingleSizeImageList = self._clone_without_tensor()
            if image_tensor.size(dim=0) > self.channel:
                tensor = self.change_channel(image_tensor.size(dim=0))._as_single_size_image_list()._tensor
            else:
                tensor = self._tensor
            if image_tensor.size(dim=0) < tensor.size(dim=1):
                if tensor.size(dim=1) == 3:
                    image_list_single._tensor = torch.cat(
                        [tensor, torch.cat([image_tensor, image_tensor, image_tensor], dim=0).unsqueeze(dim=0)],
                    )
                elif image_tensor.size(dim=0) == 1:
                    image_list_single._tensor = torch.cat(
                        [
                            tensor,
                            torch.cat(
                                [image_tensor, image_tensor, image_tensor, torch.full(image_tensor.size(), 255)],
                                dim=0,
                            ).unsqueeze(dim=0),
                        ],
                    )
                else:
                    image_list_single._tensor = torch.cat(
                        [
                            tensor,
                            torch.cat([image_tensor, torch.full(image_tensor[0:1].size(), 255)], dim=0).unsqueeze(
                                dim=0,
                            ),
                        ],
                    )
            else:
                image_list_single._tensor = torch.cat([tensor, image_tensor.unsqueeze(dim=0)])
            image_list_single._tensor_positions_to_indices.append(index)
            image_list_single._indices_to_tensor_positions[index] = len(self)

            return image_list_single
        else:
            image_list_multi: _MultiSizeImageList = _MultiSizeImageList()
            image_list_multi._image_list_dict[(self.widths[0], self.heights[0])] = self
            image_list_multi._image_list_dict[(image_tensor.size(dim=2), image_tensor.size(dim=1))] = (
                _SingleSizeImageList._create_image_list([image_tensor], [index])
            )

            if image_tensor.size(dim=0) != self.channel:
                image_list_multi = image_list_multi.change_channel(
                    max(image_tensor.size(dim=0), self.channel),
                )._as_multi_size_image_list()
            for _index in self._tensor_positions_to_indices:
                image_list_multi._indices_to_image_size_dict[_index] = (self.widths[0], self.heights[0])
            image_list_multi._indices_to_image_size_dict[index] = (image_tensor.size(dim=2), image_tensor.size(dim=1))

            return image_list_multi

    def add_images(self, images: list[Image] | ImageList) -> ImageList:
        import torch

        _init_default_device()

        from safeds.data.image.containers._empty_image_list import _EmptyImageList
        from safeds.data.image.containers._multi_size_image_list import _MultiSizeImageList

        if isinstance(images, _EmptyImageList) or isinstance(images, list) and len(images) == 0:
            return self

        first_new_index = max(self._tensor_positions_to_indices) + 1
        if isinstance(images, list):
            images = ImageList.from_images(images)
        if isinstance(images, _MultiSizeImageList):
            image_list_multi: _MultiSizeImageList = _MultiSizeImageList()
            image_list_multi._image_list_dict[(self.widths[0], self.heights[0])] = self
            for index in self._tensor_positions_to_indices:
                image_list_multi._indices_to_image_size_dict[index] = (self.widths[0], self.heights[0])
            return image_list_multi.add_images(images)
        else:
            images_as_single_size_image_list: _SingleSizeImageList = images._as_single_size_image_list()
            new_indices = [
                index + first_new_index for index in images_as_single_size_image_list._tensor_positions_to_indices
            ]
            if (
                images_as_single_size_image_list.widths[0] == self.widths[0]
                and images_as_single_size_image_list.heights[0] == self.heights[0]
            ):
                image_list_single: _SingleSizeImageList = self._clone_without_tensor()._as_single_size_image_list()
                image_list_single._tensor_positions_to_indices += new_indices
                if self.channel > images_as_single_size_image_list.channel:
                    image_list_single._tensor = torch.cat(
                        [
                            self._tensor,
                            _SingleSizeImageList._change_channel_of_tensor(
                                images_as_single_size_image_list._tensor,
                                self.channel,
                            ),
                        ],
                    )
                elif self.channel < images_as_single_size_image_list.channel:
                    image_list_single._tensor = torch.cat(
                        [
                            _SingleSizeImageList._change_channel_of_tensor(
                                self._tensor,
                                images_as_single_size_image_list.channel,
                            ),
                            images_as_single_size_image_list._tensor,
                        ],
                    )
                else:
                    image_list_single._tensor = torch.cat([self._tensor, images_as_single_size_image_list._tensor])
                image_list_single._indices_to_tensor_positions = (
                    image_list_single._calc_new_indices_to_tensor_positions()
                )
                return image_list_single
            else:
                image_list_multi = _MultiSizeImageList()
                image_list_multi._image_list_dict[(self.widths[0], self.heights[0])] = self
                image_list_multi._image_list_dict[
                    (images_as_single_size_image_list.widths[0], images_as_single_size_image_list.heights[0])
                ] = images_as_single_size_image_list
                for index in self._tensor_positions_to_indices:
                    image_list_multi._indices_to_image_size_dict[index] = (self.widths[0], self.heights[0])
                for index in images_as_single_size_image_list._tensor_positions_to_indices:
                    image_list_multi._indices_to_image_size_dict[index + first_new_index] = (
                        images_as_single_size_image_list.widths[0],
                        images_as_single_size_image_list.heights[0],
                    )

                if images_as_single_size_image_list.channel != self.channel:
                    image_list_multi = image_list_multi.change_channel(
                        max(images_as_single_size_image_list.channel, self.channel),
                    )._as_multi_size_image_list()
                return image_list_multi

    def remove_image_by_index(self, index: int | list[int]) -> ImageList:
        if isinstance(index, int):
            index = [index]

        invalid_indices = []
        for _i in index:
            if _i not in self._tensor_positions_to_indices:
                invalid_indices.append(_i)
        if len(invalid_indices) > 0:
            raise IndexOutOfBoundsError(invalid_indices)

        return self._remove_image_by_index_ignore_invalid(index)

    def _remove_image_by_index_ignore_invalid(self, index: int | list[int]) -> ImageList:
        from safeds.data.image.containers._empty_image_list import _EmptyImageList

        if isinstance(index, int):
            index = [index]

        all_indices_to_remove = True
        for i in self._tensor_positions_to_indices:
            if i not in index:
                all_indices_to_remove = False
                continue
        if all_indices_to_remove:
            return _EmptyImageList()

        image_list = _SingleSizeImageList()
        image_list._tensor = self._tensor[
            [i for i, v in enumerate(self._tensor_positions_to_indices) if v not in index]
        ].clone()
        image_list._tensor_positions_to_indices = [
            i - len([k for k in index if k < i]) for i in self._tensor_positions_to_indices if i not in index
        ]
        image_list._indices_to_tensor_positions = image_list._calc_new_indices_to_tensor_positions()
        return image_list

    def remove_images_with_size(self, width: int, height: int) -> ImageList:
        from safeds.data.image.containers._empty_image_list import _EmptyImageList

        _check_remove_images_with_size_errors(width, height)

        if self.widths[0] == width and self.heights[0] == height:
            return _EmptyImageList()
        else:
            return self

    def remove_duplicate_images(self) -> ImageList:
        image_list = _SingleSizeImageList()
        tensor_cpu_unique, new_indices = self._tensor.cpu().unique(
            dim=0,
            return_inverse=True,
        )  # Works somehow faster on cpu
        image_list._tensor = tensor_cpu_unique.to(self._tensor.device)
        indices, indices_to_remove = [], []
        offset_indices: list[int] = []
        for index_n, index_v in enumerate(new_indices):
            if list(new_indices).index(index_v) == index_n:
                indices.append(
                    self._tensor_positions_to_indices[
                        index_v + len([offset for offset in offset_indices if index_v > offset])
                    ],
                )
            else:
                offset_indices.append(index_v)
        indices_to_remove = [index_rem for index_rem in self._tensor_positions_to_indices if index_rem not in indices]
        for index_r in indices_to_remove:
            indices = [index - 1 if index > index_r else index for index in indices]
        image_list._tensor_positions_to_indices = indices
        image_list._indices_to_tensor_positions = image_list._calc_new_indices_to_tensor_positions()
        return image_list

    def shuffle_images(self) -> ImageList:
        image_list = self._clone()._as_single_size_image_list()
        random.shuffle(image_list._tensor_positions_to_indices)
        image_list._indices_to_tensor_positions = image_list._calc_new_indices_to_tensor_positions()
        return image_list

    def resize(self, new_width: int, new_height: int) -> ImageList:
        from torchvision.transforms import InterpolationMode
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        _check_resize_errors(new_width, new_height)
        image_list = self._clone_without_tensor()
        image_list._tensor = func2.resize(
            self._tensor,
            size=[new_height, new_width],
            interpolation=InterpolationMode.NEAREST,
        )
        return image_list

    def convert_to_grayscale(self) -> ImageList:
        image_list = self._clone_without_tensor()
        image_list._tensor = _SingleSizeImageList._convert_tensor_to_grayscale(self._tensor)
        return image_list

    @staticmethod
    def _convert_tensor_to_grayscale(tensor: Tensor) -> Tensor:
        import torch
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        if tensor.size(dim=-3) == 4:
            return torch.cat(
                [func2.rgb_to_grayscale(tensor[:, 0:3], num_output_channels=3), tensor[:, 3].unsqueeze(dim=1)],
                dim=1,
            )
        else:
            return func2.rgb_to_grayscale(tensor[:, 0:3], num_output_channels=tensor.size(dim=-3))

    def crop(self, x: int, y: int, width: int, height: int) -> ImageList:
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        _check_crop_errors_and_warnings(x, y, width, height, self.widths[0], self.heights[0], plural=True)
        image_list = self._clone_without_tensor()
        image_list._tensor = func2.crop(self._tensor, x, y, height, width)
        return image_list

    def flip_vertically(self) -> ImageList:
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        image_list = self._clone_without_tensor()
        image_list._tensor = func2.vertical_flip(self._tensor)
        return image_list

    def flip_horizontally(self) -> ImageList:
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        image_list = self._clone_without_tensor()
        image_list._tensor = func2.horizontal_flip(self._tensor)
        return image_list

    def adjust_brightness(self, factor: float) -> ImageList:
        import torch
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        _check_adjust_brightness_errors_and_warnings(factor, plural=True)
        image_list = self._clone_without_tensor()
        if self.channel == 4:
            image_list._tensor = torch.cat(
                [func2.adjust_brightness(self._tensor[:, 0:3], factor * 1.0), self._tensor[:, 3].unsqueeze(dim=1)],
                dim=1,
            )
        else:
            image_list._tensor = func2.adjust_brightness(self._tensor, factor * 1.0)
        return image_list

    def add_noise(self, standard_deviation: float) -> ImageList:
        import torch

        _init_default_device()

        _check_add_noise_errors(standard_deviation)
        image_list = self._clone_without_tensor()
        image_list._tensor = (
            self._tensor + torch.normal(0, standard_deviation, self._tensor.size()).to(_get_device()) * 255
        )
        return image_list

    def adjust_contrast(self, factor: float) -> ImageList:
        import torch
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        _check_adjust_contrast_errors_and_warnings(factor, plural=True)
        image_list = self._clone_without_tensor()
        if self.channel == 4:
            image_list._tensor = torch.cat(
                [func2.adjust_contrast(self._tensor[:, 0:3], factor * 1.0), self._tensor[:, 3].unsqueeze(dim=1)],
                dim=1,
            )
        else:
            image_list._tensor = func2.adjust_contrast(self._tensor, factor * 1.0)
        return image_list

    def adjust_color_balance(self, factor: float) -> ImageList:
        _check_adjust_color_balance_errors_and_warnings(factor, self.channel, plural=True)
        image_list = self._clone_without_tensor()
        image_list._tensor = self.convert_to_grayscale()._as_single_size_image_list()._tensor * (
            1.0 - factor * 1.0
        ) + self._tensor * (factor * 1.0)
        return image_list

    def blur(self, radius: int) -> ImageList:
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        _check_blur_errors_and_warnings(radius, min(self.widths[0], self.heights[0]), plural=True)
        image_list = self._clone_without_tensor()
        image_list._tensor = func2.gaussian_blur(self._tensor, [radius * 2 + 1, radius * 2 + 1])
        return image_list

    def sharpen(self, factor: float) -> ImageList:
        import torch
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        _check_sharpen_errors_and_warnings(factor, plural=True)
        image_list = self._clone_without_tensor()
        if self.channel == 4:
            image_list._tensor = torch.cat(
                [func2.adjust_sharpness(self._tensor[:, 0:3], factor * 1.0), self._tensor[:, 3].unsqueeze(dim=1)],
                dim=1,
            )
        else:
            image_list._tensor = func2.adjust_sharpness(self._tensor, factor * 1.0)
        return image_list

    def invert_colors(self) -> ImageList:
        import torch
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        image_list = self._clone_without_tensor()
        if self.channel == 4:
            image_list._tensor = torch.cat(
                [func2.invert(self._tensor[:, 0:3]), self._tensor[:, 3].unsqueeze(dim=1)],
                dim=1,
            )
        else:
            image_list._tensor = func2.invert(self._tensor)
        return image_list

    def rotate_right(self) -> ImageList:
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        image_list = self._clone_without_tensor()
        image_list._tensor = func2.rotate(self._tensor, -90, expand=True)
        return image_list

    def rotate_left(self) -> ImageList:
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        image_list = self._clone_without_tensor()
        image_list._tensor = func2.rotate(self._tensor, 90, expand=True)
        return image_list

    def find_edges(self) -> ImageList:
        import torch

        _init_default_device()

        kernel = Image._filter_edges_kernel()
        edges_tensor = torch.clamp(
            torch.nn.functional.conv2d(
                self.convert_to_grayscale()._as_single_size_image_list()._tensor.float()[:, 0].unsqueeze(dim=1),
                kernel,
                padding="same",
            ),
            0,
            255,
        ).to(torch.uint8)
        image_list = self._clone_without_tensor()
        if self.channel == 3:
            image_list._tensor = edges_tensor.repeat(1, 3, 1, 1)
        elif self.channel == 4:
            image_list._tensor = torch.cat(
                [edges_tensor.repeat(1, 3, 1, 1), self._tensor[:, 3].unsqueeze(dim=1)],
                dim=1,
            )
        else:
            image_list._tensor = edges_tensor
        return image_list



from __future__ import annotations

from abc import ABC, abstractmethod

from safeds._utils import _structural_hash
from safeds._validation import _check_bounds, _ClosedBound


class _RandomForestBase(ABC):
    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    @abstractmethod
    def __init__(
        self,
        number_of_trees: int,
        maximum_depth: int | None,
        minimum_number_of_samples_in_leaves: int,
    ) -> None:
        # Validation
        _check_bounds("number_of_trees", number_of_trees, lower_bound=_ClosedBound(1))
        _check_bounds("maximum_depth", maximum_depth, lower_bound=_ClosedBound(1))
        _check_bounds(
            "minimum_number_of_samples_in_leaves",
            minimum_number_of_samples_in_leaves,
            lower_bound=_ClosedBound(1),
        )

        # Hyperparameters
        self._number_of_trees: int = number_of_trees
        self._maximum_depth: int | None = maximum_depth
        self._minimum_number_of_samples_in_leaves: int = minimum_number_of_samples_in_leaves

    def __hash__(self) -> int:
        return _structural_hash(
            self._number_of_trees,
            self._maximum_depth,
            self._minimum_number_of_samples_in_leaves,
        )

    # ------------------------------------------------------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def number_of_trees(self) -> int:
        """The number of trees used in the random forest."""
        return self._number_of_trees

    @property
    def maximum_depth(self) -> int | None:
        """The maximum depth of each tree."""
        return self._maximum_depth

    @property
    def minimum_number_of_samples_in_leaves(self) -> int:
        """The minimum number of samples that must remain in the leaves of each tree."""
        return self._minimum_number_of_samples_in_leaves



from __future__ import annotations

from typing import TYPE_CHECKING

from safeds._utils import _figure_to_image
from safeds.exceptions import NonNumericColumnError

if TYPE_CHECKING:
    from safeds.data.image.containers import Image
    from safeds.data.tabular.containers import Column


class ColumnPlotter:
    """
    A class that contains plotting methods for a column.

    Parameters
    ----------
    column:
        The column to plot.

    Examples
    --------
    >>> from safeds.data.tabular.containers import Column
    >>> column = Column("test", [1, 2, 3])
    >>> plotter = column.plot
    """

    def __init__(self, column: Column):
        self._column: Column = column

    def box_plot(self) -> Image:
        """
        Create a box plot for the values in the column. This is only possible for numeric columns.

        Returns
        -------
        box_plot:
            The box plot as an image.

        Raises
        ------
        TypeError
            If the column is not numeric.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> boxplot = column.plot.box_plot()
        """
        if self._column.number_of_rows > 0 and not self._column.is_numeric:
            # TODO better error message
            raise NonNumericColumnError(f"{self._column.name} is of type {self._column.type}.")

        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()
        ax.boxplot(
            self._column._series.drop_nulls(),
            patch_artist=True,
        )

        ax.set(title=self._column.name)
        ax.set_xticks([])
        ax.yaxis.grid(visible=True)
        fig.tight_layout()

        return _figure_to_image(fig)

    def histogram(self, *, maximum_number_of_bins: int = 10) -> Image:
        """
        Create a histogram for the values in the column.

        Parameters
        ----------
        maximum_number_of_bins:
            The maximum number of bins to use in the histogram. Default is 10.

        Returns
        -------
        histogram:
            The plot as an image.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> histogram = column.plot.histogram()
        """
        return self._column.to_table().plot.histograms(maximum_number_of_bins=maximum_number_of_bins)

    def lag_plot(self, lag: int) -> Image:
        """
        Create a lag plot for the values in the column.

        Parameters
        ----------
        lag:
            The amount of lag.

        Returns
        -------
        lag_plot:
            The plot as an image.

        Raises
        ------
        TypeError
            If the column is not numeric.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("values", [1, 2, 3, 4])
        >>> image = column.plot.lag_plot(2)
        """
        if self._column.number_of_rows > 0 and not self._column.is_numeric:
            # TODO better error message
            raise NonNumericColumnError("This time series target contains non-numerical columns.")

        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()
        series = self._column._series
        ax.scatter(
            x=series.slice(0, max(len(self._column) - lag, 0)),
            y=series.slice(lag),
        )
        ax.set(
            xlabel="y(t)",
            ylabel=f"y(t + {lag})",
        )
        fig.tight_layout()

        return _figure_to_image(fig)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Generic, TypeVar

if TYPE_CHECKING:
    import polars as pl

T_co = TypeVar("T_co", covariant=True)
P = TypeVar("P")
R_co = TypeVar("R_co", covariant=True)


class Cell(ABC, Generic[T_co]):
    """
    A single value in a table.

    This class cannot be instantiated directly. It is only used for arguments of callbacks.
    """

    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    # "Boolean" operators (actually bitwise) -----------------------------------

    @abstractmethod
    def __invert__(self) -> Cell[bool]: ...

    @abstractmethod
    def __and__(self, other: bool | Cell[bool]) -> Cell[bool]: ...

    @abstractmethod
    def __rand__(self, other: bool | Cell[bool]) -> Cell[bool]: ...

    @abstractmethod
    def __or__(self, other: bool | Cell[bool]) -> Cell[bool]: ...

    @abstractmethod
    def __ror__(self, other: bool | Cell[bool]) -> Cell[bool]: ...

    @abstractmethod
    def __xor__(self, other: bool | Cell[bool]) -> Cell[bool]: ...

    @abstractmethod
    def __rxor__(self, other: bool | Cell[bool]) -> Cell[bool]: ...

    # Comparison ---------------------------------------------------------------

    @abstractmethod
    def __eq__(self, other: object) -> Cell[bool]:  # type: ignore[override]
        ...

    @abstractmethod
    def __ge__(self, other: Any) -> Cell[bool]: ...

    @abstractmethod
    def __gt__(self, other: Any) -> Cell[bool]: ...

    @abstractmethod
    def __le__(self, other: Any) -> Cell[bool]: ...

    @abstractmethod
    def __lt__(self, other: Any) -> Cell[bool]: ...

    @abstractmethod
    def __ne__(self, other: object) -> Cell[bool]:  # type: ignore[override]
        ...

    # Numeric operators --------------------------------------------------------

    @abstractmethod
    def __abs__(self) -> Cell[R_co]: ...

    @abstractmethod
    def __ceil__(self) -> Cell[R_co]: ...

    @abstractmethod
    def __floor__(self) -> Cell[R_co]: ...

    @abstractmethod
    def __neg__(self) -> Cell[R_co]: ...

    @abstractmethod
    def __pos__(self) -> Cell[R_co]: ...

    @abstractmethod
    def __add__(self, other: Any) -> Cell[R_co]: ...

    @abstractmethod
    def __radd__(self, other: Any) -> Cell[R_co]: ...

    @abstractmethod
    def __floordiv__(self, other: Any) -> Cell[R_co]: ...

    @abstractmethod
    def __rfloordiv__(self, other: Any) -> Cell[R_co]: ...

    @abstractmethod
    def __mod__(self, other: Any) -> Cell[R_co]: ...

    @abstractmethod
    def __rmod__(self, other: Any) -> Cell[R_co]: ...

    @abstractmethod
    def __mul__(self, other: Any) -> Cell[R_co]: ...

    @abstractmethod
    def __rmul__(self, other: Any) -> Cell[R_co]: ...

    @abstractmethod
    def __pow__(self, other: float | Cell[P]) -> Cell[R_co]: ...

    @abstractmethod
    def __rpow__(self, other: float | Cell[P]) -> Cell[R_co]: ...

    @abstractmethod
    def __sub__(self, other: Any) -> Cell[R_co]: ...

    @abstractmethod
    def __rsub__(self, other: Any) -> Cell[R_co]: ...

    @abstractmethod
    def __truediv__(self, other: Any) -> Cell[R_co]: ...

    @abstractmethod
    def __rtruediv__(self, other: Any) -> Cell[R_co]: ...

    # Other --------------------------------------------------------------------

    @abstractmethod
    def __hash__(self) -> int: ...

    @abstractmethod
    def __sizeof__(self) -> int: ...

    # ------------------------------------------------------------------------------------------------------------------
    # Boolean operations
    # ------------------------------------------------------------------------------------------------------------------

    def not_(self) -> Cell[bool]:
        """
        Negate a boolean. This is equivalent to the `~` operator.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("example", [True, False])
        >>> column.transform(lambda cell: cell.not_())
        +---------+
        | example |
        | ---     |
        | bool    |
        +=========+
        | false   |
        | true    |
        +---------+

        >>> column.transform(lambda cell: ~cell)
        +---------+
        | example |
        | ---     |
        | bool    |
        +=========+
        | false   |
        | true    |
        +---------+
        """
        return self.__invert__()

    def and_(self, other: bool | Cell[bool]) -> Cell[bool]:
        """
        Perform a boolean AND operation. This is equivalent to the `&` operator.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("example", [True, False])
        >>> column.transform(lambda cell: cell.and_(False))
        +---------+
        | example |
        | ---     |
        | bool    |
        +=========+
        | false   |
        | false   |
        +---------+

        >>> column.transform(lambda cell: cell & False)
        +---------+
        | example |
        | ---     |
        | bool    |
        +=========+
        | false   |
        | false   |
        +---------+
        """
        return self.__and__(other)

    def or_(self, other: bool | Cell[bool]) -> Cell[bool]:
        """
        Perform a boolean OR operation. This is equivalent to the `|` operator.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("example", [True, False])
        >>> column.transform(lambda cell: cell.or_(True))
        +---------+
        | example |
        | ---     |
        | bool    |
        +=========+
        | true    |
        | true    |
        +---------+

        >>> column.transform(lambda cell: cell | True)
        +---------+
        | example |
        | ---     |
        | bool    |
        +=========+
        | true    |
        | true    |
        +---------+
        """
        return self.__or__(other)

    def xor(self, other: bool | Cell[bool]) -> Cell[bool]:
        """
        Perform a boolean XOR operation. This is equivalent to the `^` operator.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("example", [True, False])
        >>> column.transform(lambda cell: cell.xor(True))
        +---------+
        | example |
        | ---     |
        | bool    |
        +=========+
        | false   |
        | true    |
        +---------+

        >>> column.transform(lambda cell: cell ^ True)
        +---------+
        | example |
        | ---     |
        | bool    |
        +=========+
        | false   |
        | true    |
        +---------+
        """
        return self.__xor__(other)

    # ------------------------------------------------------------------------------------------------------------------
    # Numeric operations
    # ------------------------------------------------------------------------------------------------------------------

    def abs(self) -> Cell[R_co]:
        """
        Get the absolute value.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("example", [1, -2])
        >>> column.transform(lambda cell: cell.abs())
        +---------+
        | example |
        |     --- |
        |     i64 |
        +=========+
        |       1 |
        |       2 |
        +---------+
        """
        return self.__abs__()

    def ceil(self) -> Cell[R_co]:
        """
        Round up to the nearest integer.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("example", [1.1, 2.9])
        >>> column.transform(lambda cell: cell.ceil())
        +---------+
        | example |
        |     --- |
        |     f64 |
        +=========+
        | 2.00000 |
        | 3.00000 |
        +---------+
        """
        return self.__ceil__()

    def floor(self) -> Cell[R_co]:
        """
        Round down to the nearest integer.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("example", [1.1, 2.9])
        >>> column.transform(lambda cell: cell.floor())
        +---------+
        | example |
        |     --- |
        |     f64 |
        +=========+
        | 1.00000 |
        | 2.00000 |
        +---------+
        """
        return self.__floor__()

    def neg(self) -> Cell[R_co]:
        """
        Negate the value.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("example", [1, -2])
        >>> column.transform(lambda cell: cell.neg())
        +---------+
        | example |
        |     --- |
        |     i64 |
        +=========+
        |      -1 |
        |       2 |
        +---------+
        """
        return self.__neg__()

    def add(self, other: Any) -> Cell[R_co]:
        """
        Add a value. This is equivalent to the `+` operator.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("example", [1, 2])
        >>> column.transform(lambda cell: cell.add(3))
        +---------+
        | example |
        |     --- |
        |     i64 |
        +=========+
        |       4 |
        |       5 |
        +---------+

        >>> column.transform(lambda cell: cell + 3)
        +---------+
        | example |
        |     --- |
        |     i64 |
        +=========+
        |       4 |
        |       5 |
        +---------+
        """
        return self.__add__(other)

    def mod(self, other: Any) -> Cell[R_co]:
        """
        Perform a modulo operation. This is equivalent to the `%` operator.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("example", [5, 6])
        >>> column.transform(lambda cell: cell.mod(3))
        +---------+
        | example |
        |     --- |
        |     i64 |
        +=========+
        |       2 |
        |       0 |
        +---------+

        >>> column.transform(lambda cell: cell % 3)
        +---------+
        | example |
        |     --- |
        |     i64 |
        +=========+
        |       2 |
        |       0 |
        +---------+
        """
        return self.__mod__(other)

    def mul(self, other: Any) -> Cell[R_co]:
        """
        Multiply by a value. This is equivalent to the `*` operator.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("example", [2, 3])
        >>> column.transform(lambda cell: cell.mul(4))
        +---------+
        | example |
        |     --- |
        |     i64 |
        +=========+
        |       8 |
        |      12 |
        +---------+

        >>> column.transform(lambda cell: cell * 4)
        +---------+
        | example |
        |     --- |
        |     i64 |
        +=========+
        |       8 |
        |      12 |
        +---------+
        """
        return self.__mul__(other)

    def pow(self, other: float | Cell[P]) -> Cell[R_co]:
        """
        Raise to a power. This is equivalent to the `**` operator.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("example", [2, 3])
        >>> column.transform(lambda cell: cell.pow(3))
        +---------+
        | example |
        |     --- |
        |     i64 |
        +=========+
        |       8 |
        |      27 |
        +---------+

        >>> column.transform(lambda cell: cell ** 3)
        +---------+
        | example |
        |     --- |
        |     i64 |
        +=========+
        |       8 |
        |      27 |
        +---------+
        """
        return self.__pow__(other)

    def sub(self, other: Any) -> Cell[R_co]:
        """
        Subtract a value. This is equivalent to the `-` operator.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("example", [5, 6])
        >>> column.transform(lambda cell: cell.sub(3))
        +---------+
        | example |
        |     --- |
        |     i64 |
        +=========+
        |       2 |
        |       3 |
        +---------+

        >>> column.transform(lambda cell: cell - 3)
        +---------+
        | example |
        |     --- |
        |     i64 |
        +=========+
        |       2 |
        |       3 |
        +---------+
        """
        return self.__sub__(other)

    def div(self, other: Any) -> Cell[R_co]:
        """
        Divide by a value. This is equivalent to the `/` operator.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("example", [6, 8])
        >>> column.transform(lambda cell: cell.div(2))
        +---------+
        | example |
        |     --- |
        |     f64 |
        +=========+
        | 3.00000 |
        | 4.00000 |
        +---------+

        >>> column.transform(lambda cell: cell / 2)
        +---------+
        | example |
        |     --- |
        |     f64 |
        +=========+
        | 3.00000 |
        | 4.00000 |
        +---------+
        """
        return self.__truediv__(other)

    # ------------------------------------------------------------------------------------------------------------------
    # Comparison operations
    # ------------------------------------------------------------------------------------------------------------------

    def eq(self, other: Any) -> Cell[bool]:
        """
        Check if equal to a value. This is equivalent to the `==` operator.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("example", [1, 2])
        >>> column.transform(lambda cell: cell.eq(2))
        +---------+
        | example |
        | ---     |
        | bool    |
        +=========+
        | false   |
        | true    |
        +---------+

        >>> column.transform(lambda cell: cell == 2)
        +---------+
        | example |
        | ---     |
        | bool    |
        +=========+
        | false   |
        | true    |
        +---------+
        """
        return self.__eq__(other)

    def ge(self, other: Any) -> Cell[bool]:
        """
        Check if greater than or equal to a value. This is equivalent to the `>=` operator.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("example", [1, 2])
        >>> column.transform(lambda cell: cell.ge(2))
        +---------+
        | example |
        | ---     |
        | bool    |
        +=========+
        | false   |
        | true    |
        +---------+

        >>> column.transform(lambda cell: cell >= 2)
        +---------+
        | example |
        | ---     |
        | bool    |
        +=========+
        | false   |
        | true    |
        +---------+
        """
        return self.__ge__(other)

    def gt(self, other: Any) -> Cell[bool]:
        """
        Check if greater than a value. This is equivalent to the `>` operator.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("example", [1, 2])
        >>> column.transform(lambda cell: cell.gt(2))
        +---------+
        | example |
        | ---     |
        | bool    |
        +=========+
        | false   |
        | false   |
        +---------+

        >>> column.transform(lambda cell: cell > 2)
        +---------+
        | example |
        | ---     |
        | bool    |
        +=========+
        | false   |
        | false   |
        +---------+
        """
        return self.__gt__(other)

    def le(self, other: Any) -> Cell[bool]:
        """
        Check if less than or equal to a value. This is equivalent to the `<=` operator.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("example", [1, 2])
        >>> column.transform(lambda cell: cell.le(2))
        +---------+
        | example |
        | ---     |
        | bool    |
        +=========+
        | true    |
        | true    |
        +---------+

        >>> column.transform(lambda cell: cell <= 2)
        +---------+
        | example |
        | ---     |
        | bool    |
        +=========+
        | true    |
        | true    |
        +---------+
        """
        return self.__le__(other)

    def lt(self, other: Any) -> Cell[bool]:
        """
        Check if less than a value. This is equivalent to the `<` operator.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("example", [1, 2])
        >>> column.transform(lambda cell: cell.lt(2))
        +---------+
        | example |
        | ---     |
        | bool    |
        +=========+
        | true    |
        | false   |
        +---------+

        >>> column.transform(lambda cell: cell < 2)
        +---------+
        | example |
        | ---     |
        | bool    |
        +=========+
        | true    |
        | false   |
        +---------+
        """
        return self.__lt__(other)

    # ------------------------------------------------------------------------------------------------------------------
    # Internal
    # ------------------------------------------------------------------------------------------------------------------

    @property
    @abstractmethod
    def _polars_expression(self) -> pl.Expr:
        """The Polars expression that corresponds to this cell."""

    @abstractmethod
    def _equals(self, other: object) -> bool:
        """
        Check if this cell is equal to another object.

        This method is needed because the `__eq__` method is used for element-wise comparisons.
        """



from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any

from safeds._config import _init_default_device
from safeds._utils import _structural_hash

from ._layer import Layer

if TYPE_CHECKING:
    from torch import Tensor, nn

    from safeds.data.image.typing import ImageSize


def _create_internal_model() -> nn.Module:
    from torch import nn

    _init_default_device()

    class _InternalLayer(nn.Module):
        def __init__(self) -> None:
            super().__init__()
            self._layer = nn.Flatten()

        def forward(self, x: Tensor) -> Tensor:
            return self._layer(x)

    return _InternalLayer()


class FlattenLayer(Layer):
    def __init__(self) -> None:
        """Create a Flatten Layer."""
        self._input_size: ImageSize | None = None
        self._output_size: int | None = None

    def _get_internal_layer(self, **kwargs: Any) -> nn.Module:  # noqa: ARG002
        return _create_internal_model()

    @property
    def input_size(self) -> ImageSize:
        """
        Get the input_size of this layer.

        Returns
        -------
        result :
            The amount of values being passed into this layer.

        Raises
        ------
        ValueError
            If the input_size is not yet set
        """
        if self._input_size is None:
            raise ValueError("The input_size is not yet set.")
        return self._input_size

    @property
    def output_size(self) -> int:
        """
        Get the output_size of this layer.

        Returns
        -------
        result :
            The Number of Neurons in this layer.

        Raises
        ------
        ValueError
            If the input_size is not yet set
        """
        if self._input_size is None:
            raise ValueError(
                "The input_size is not yet set. The layer cannot compute the output_size if the input_size is not set.",
            )
        if self._output_size is None:
            self._output_size = self._input_size.width * self._input_size.height * self._input_size.channel
        return self._output_size

    def _set_input_size(self, input_size: int | ImageSize) -> None:
        if isinstance(input_size, int):
            raise TypeError("The input_size of a flatten layer has to be of type ImageSize.")
        self._input_size = input_size
        self._output_size = None

    def __hash__(self) -> int:
        """
        Return a deterministic hash value for this flatten layer.

        Returns
        -------
        hash:
            the hash value
        """
        return _structural_hash(self._input_size, self._output_size)

    def __eq__(self, other: object) -> bool:
        """
        Compare two flatten layer.

        Parameters
        ----------
        other:
            The flatten layer to compare to.

        Returns
        -------
        equals:
            Whether the two flatten layer are the same.
        """
        if not isinstance(other, FlattenLayer):
            return NotImplemented
        return (self is other) or (self._input_size == other._input_size and self._output_size == other._output_size)

    def __sizeof__(self) -> int:
        """
        Return the complete size of this object.

        Returns
        -------
        size:
            Size of this object in bytes.
        """
        return sys.getsizeof(self._input_size) + sys.getsizeof(self._output_size)
